"""
f2py wrapper

Patches the default behaviour of f2py with some custom error handling.
This was inspired by the amazing [f90wrap](https://github.com/jameskermode/f90wrap).
"""
import sys

import numpy.f2py.rules as f2py_rules
from numpy.f2py import main as f2py_main  # type: ignore


def patch_f2py() -> None:
    """
    Patch f2py to raise a RuntimeError if :c:func:`on_error` is called during execution

    Interrupt signals are also captured and handled.
    This has only been tested with numpy 1.21.0
    and may be flakey if f2py has substantial changes.
    """
    includes_inject = "#includes0#\n"

    if sys.platform == "win32":
        includes_inject = (
            includes_inject + "#include <signal.h>\n#include <setjmpex.h>\n"
        )
    else:
        includes_inject = includes_inject + "#include <setjmp.h>\n"

    includes_inject = (
        includes_inject
        + """
    #include <stdlib.h>
    #include <string.h>

    #define ABORT_BUFFER_SIZE 512
    extern jmp_buf environment_buffer;
    extern char abort_message[ABORT_BUFFER_SIZE];
    void on_error_(char *message, size_t len);
    void abort_int_handler(int signum);

    jmp_buf environment_buffer;
    char abort_message[ABORT_BUFFER_SIZE];

    // Error handler for fgen
    void on_error_(char *message, size_t len_message)
    {
      strncpy(abort_message, message, ABORT_BUFFER_SIZE);
      abort_message[ABORT_BUFFER_SIZE-1] = '\\0';
      longjmp(environment_buffer, 0);
    }

    // Handle sigint signals (ctrl + c) during a f2py calls
    void sigint_handler(int signum)
    {
      char message[] = "User interrupt occurred";
      on_error_(message, strlen(message));
    }
    """
    )

    f2py_rules.module_rules["modulebody"] = f2py_rules.module_rules[
        "modulebody"
    ].replace("#includes0#\n", includes_inject)

    f2py_rules.routine_rules["body"] = f2py_rules.routine_rules["body"].replace(  # type: ignore
        "volatile int f2py_success = 1;\n",
        "volatile int f2py_success = 1; int setjmp_value;",
    )

    f2py_rules.routine_rules["body"] = f2py_rules.routine_rules["body"].replace(  # type: ignore
        "#callfortranroutine#\n",
        """
        PyOS_sighandler_t _npy_sig_save;

        // Catch any sigint (ctrl + c) with a call to sigint_handler
        _npy_sig_save = PyOS_setsig(SIGINT, sigint_handler);
        setjmp_value = setjmp(environment_buffer);

        if (setjmp_value != 0) {
          // jumped back as a result of a call to longjmp
          // Raise a RuntimeError
          PyOS_setsig(SIGINT, _npy_sig_save);
          PyErr_SetString(PyExc_RuntimeError, abort_message);
        } else {
         #callfortranroutine#
         PyOS_setsig(SIGINT, _npy_sig_save);
        }
        """,
    )


if __name__ == "__main__":
    patch_f2py()
    f2py_main()



"""
Data model of a value, excluding units
"""

from __future__ import annotations

from typing import Any

from attrs import define, field, validators

from fgen.fortran_parsing import FortranDataType

_NUMERIC_TYPE_STRINGS: list[str] = ["float", "int"]
"""
Strings that can appear in Python types that indicate that the Python type is numeric
"""


@define
class UnitlessValue:
    """
    Data model of a value excluding units

    This defines the value's Fortran data type and other metadata.
    It also allows us to get the equivalent Python type.
    """

    name: str
    """Name of the value"""

    description: str
    """Description of the value"""

    fortran_type: str = field(validator=[validators.instance_of(str)])
    """
    The value's fortran type

    The following built-in Fortran types are supported:

    * integer
    * real
    * real(8)
    * character

    Some additional attributes are supported including:

    * fixed length dimensions
    * automatic explicit length dimensions
      (e.g. using a variable "n" to specify the size of a dimension)
    """

    is_fortran_units_holder: bool = False
    """
    Is this value the attribute which specifies the units in the Fortran derived type?

    If ``True``, it is not not included in Python wrappers,
    instead being passed automatically to Fortran
    based on the units extracted from :mod:`pint` quantities in Python.
    """

    expose_getter_to_python: bool = True
    """
    Should a getter for this value be exposed to Python?

    If ``True``, we expose a getter for this value to Python, otherwise we don't.
    """

    expose_setter_to_python: bool = False
    """
    Should a setter for this value be exposed to Python?

    If ``True``, we expose a setter for this value to Python, otherwise we don't.
    """

    @fortran_type.validator
    def _check_fortran_type(self, attribute: Any, value: str) -> None:
        try:
            FortranDataType.from_str(value)
        except Exception as exc:
            raise ValueError(  # noqa: TRY003
                f"Unsupported fortran type: {value}"
            ) from exc

    @property
    def is_numeric_type(self) -> bool:
        """
        Whether this definition defines a numeric type or not

        Returns
        -------
            ``True`` if this definition defines a numeric type, ``False`` otherwise.
        """
        pt = self.python_type_as_str()
        return any(t in pt for t in _NUMERIC_TYPE_STRINGS)

    @property
    def is_python_str_type(self) -> bool:
        """
        Whether this definition corresponds to a string

        The concept of a string obviously only applies in Python.
        In Fortran, the equivalent is a character type.

        Returns
        -------
            ``True`` if this definition defines a type
            whose Python equivalent is string,
            ``False`` otherwise.
        """
        pt = self.python_type_as_str()

        return pt == "str"

    @property
    def is_python_bool_type(self) -> bool:
        """
        Whether this definition corresponds to a boolean

        The concept of a boolean obviously only applies in Python.
        In Fortran, the equivalent is a logical type.

        Returns
        -------
            ``True`` if this definition defines a type
            whose Python equivalent is boolean,
            ``False`` otherwise.
        """
        pt = self.python_type_as_str()
        return pt == "bool"

    def as_fortran_data_type(self) -> FortranDataType:
        """
        Convert ``self`` to its equivalent ``FortranDataType``
        """
        return FortranDataType.from_str(self.fortran_type)

    def python_type_as_str(self) -> str:
        """
        Determine the string representation of the Python type

        The python templator uses this function to translate a Python
        type to a string so that it can correctly be read by Python.
        If these type hints are converted to a string naively, the
        corresponding string will not be able to be read by Python as a valid type hint.

        Returns
        -------
            String representation of :attr:`python_type`
        """
        return self.as_fortran_data_type().equivalent_python_type



"""
Wrapping strategy for enum types

The enum values are passed as integers across the Python-Fortran interface.
"""
from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Optional, Union

from attrs import define

from fgen.data_models import Method, Module, MultiReturn, Value
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
)
from fgen.wrapping_strategies.information_injection import (
    inject_wrapping_strategy_information,
)
from fgen.wrapping_strategies.passing_to_fortran_steps import (
    PassingToFortranSteps,
)
from fgen.wrapping_strategies.receiving_from_python_steps import (
    ReceivingFromPythonSteps,
)


@define
class WrappingStrategyEnum:
    """
    Wrapping strategy for enums
    """

    int_value_suffix: str = "_int_value"
    """
    Suffix to add to the variable name to indicate that this is the integer equivalent
    """

    def get_fortran_wrapper_module_getter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iget_{value.definition.name}"

    def get_fortran_wrapper_module_setter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iset_{value.definition.name}"

    def get_fortran_wrapper_statement_declarations_getters_and_setters(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> list[str]:
        return [
            f"public :: {method_name}"
            for method_name in (
                self.get_fortran_wrapper_module_getter(value),
                self.get_fortran_wrapper_module_setter(value),
            )
        ]

    def get_fortran_wrapper_statement_declarations_for_returning_method_result_to_python(  # noqa: E501, D102
        self, method: Method, prefix: str
    ) -> list[str]:
        return [f"public :: {prefix}{method.name}"]

    def get_receiving_from_python_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str] = None,
        providing_module: Optional[Module] = None,
        get_instance_callable_name: Optional[str] = None,
    ) -> ReceivingFromPythonSteps:
        return ReceivingFromPythonSteps(
            postprocessing_fortran_calls=None,
            fortran_wrapper_callable_parameters=(
                (
                    value.definition.name,
                    value.definition.as_fortran_data_type().fortran_type_attribute_declaration,
                    value,
                ),
            ),
            fortran_module_callable_args=(value,),
        )

    def get_type_attribute_declaration_for_value_to_pass_to_original_fortran_module(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return (
            value.definition.as_fortran_data_type().fortran_type_attribute_declaration
        )

    def get_fortran_for_getter(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        value_manager_get_free_instance_number: Optional[str] = None,
        value_manager_get_instance: Optional[str] = None,
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-getter.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            return_value=value,
            fortran_wrapper_module_callable=self.get_fortran_wrapper_module_getter(
                value
            ),
            class_being_wrapped=class_being_wrapped,
            fortran_module_attribute=value.definition.name,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="!"
        )

    def get_fortran_for_method_returning_wrapped_type(  # noqa: D102, PLR0913
        self,
        fortran_wrapper_module_callable: str,
        receiving_from_python_steps: ReceivingFromPythonSteps,
        return_value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        method_name: str,
        return_value_manager_get_free_instance_number: Optional[str],
        return_value_manager_get_instance: Optional[str],
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-method-returning-wrapped-type.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        input_params = (
            receiving_from_python_steps.fortran_wrapper_callable_parameter_names
        )

        result = template.render(
            fortran_wrapper_module_callable=fortran_wrapper_module_callable,
            return_value=return_value,
            input_params=input_params,
            receiving_from_python_steps=receiving_from_python_steps,
            class_being_wrapped=class_being_wrapped,
            method_name=method_name,
        )

        return inject_wrapping_strategy_information(
            result, value=return_value, wrapping_strategy=self, comment_character="!"
        )

    def get_python_user_facing_name(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.name

    def get_python_post_verify_units_input_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_return_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_argument_declaration_type_annotation(  # noqa: D102
        self, value: Value
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_getter_docstring(self, value: Value) -> str:  # noqa: D102
        template = get_template_in_directory(
            "python-getter-docstring.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(attribute=value)

        return result

    def get_python_setter_docstring(self, value: Value) -> str:  # noqa: D102
        return f"Setter for ``{value.definition.name}``"

    def generate_python_for_fortran_return_value_processing(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        fortran_module_callable: str,
        fortran_module_callable_args: Iterable[str],
        dynamic_unit: Optional[str],
    ) -> str:
        return_type = self.get_python_return_type_annotation(value)

        template = get_template_in_directory(
            "python-for-fortran-return-value-processing.py.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            result_name=self.get_python_user_facing_name(value),
            fortran_module_callable=fortran_module_callable,
            fortran_module_callable_args=fortran_module_callable_args,
            return_type=return_type,
            int_value_suffix=self.int_value_suffix,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="#"
        )

    def get_passing_to_fortran_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str],
    ) -> PassingToFortranSteps:
        argument_int_value = f"{value.definition.name}{self.int_value_suffix}"

        template = get_template_in_directory(
            "python-for-preparing-to-pass-to-fortran.py.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )
        preparatory_python_calls = template.render(
            argument_user_facing=self.get_python_user_facing_name(value),
            argument_int_value=argument_int_value,
        )

        return PassingToFortranSteps(
            preparatory_python_calls=inject_wrapping_strategy_information(
                preparatory_python_calls,
                value=value,
                wrapping_strategy=self,
                comment_character="#",
            ),
            fortran_module_callable_args=((value.definition.name, argument_int_value),),
        )



"""
Data model of a return value that contains multiple values
"""
from __future__ import annotations

from typing import Any, Optional, Union

from attrs import define, field

from fgen.data_models.unitless_value import UnitlessValue


@define
class MultiReturn:  # type: ignore # mypy confused by converter
    """
    Data model for a return type that returns more than one value

    This defines the returned values' units, Fortran data type and other metadata.
    It is the combination of a :obj:`UnitlessValueDefinition`
    and information about the units of each returned value.
    """

    definition: UnitlessValue
    """Definition of the value's key information"""

    unit: Optional[tuple[str, ...]] = field(  # type: ignore # mypy confused by converter
        default=None,
        converter=lambda x: tuple(x) if x is not None and not isinstance(x, str) else x,
    )
    """
    Units of the returned values

    The units must be able to parsed by `pint`
    and be present in the :obj:`pint.UnitRegistry`
    being used by the application (normal rules for pint).
    Some examples include: "kg", "1 / month".

    A unit is required for all numeric-types (i.e. integer, real, complex).
    For non-numeric values, the unit is unused.
    """

    dynamic_unit: Union[bool, str] = False
    """
    Whether the units should be inferred dynamically, rather than statically

    Support for this is currently not implemented.
    If you require this feature, please raise an issue.
    """

    @unit.validator
    def _check_unit(self, attribute: Any, value: tuple[str, ...] | None) -> None:
        if self.dynamic_unit:
            raise NotImplementedError()

        if self.definition.is_numeric_type:
            if value is None:
                raise ValueError(  # noqa: TRY003
                    f"A unit is required for: {self.definition.name}"
                )

            if not isinstance(value, tuple) or not all(
                isinstance(v, str) for v in value
            ):
                raise TypeError(  # noqa: TRY003
                    f"The unit for {self.definition.name} must be a tuple of strings, "
                    f"received: {value!r}"
                )

            dimension_attribute_spec = (
                self.definition.as_fortran_data_type().dimension_attribute_specification
            )
            if dimension_attribute_spec is None:
                raise ValueError(  # noqa: TRY003
                    "The Fortran method must return "
                    "more than one value for this to work. "
                    f"{self.definition.fortran_type=}"
                )

            dimensions = dimension_attribute_spec.dimensions
            if len(dimensions) > 1:
                raise NotImplementedError(
                    "Not sure which way to put the dimensions/units to make this work"
                )

            if len(value) != dimensions[0]:
                raise ValueError(  # noqa: TRY003
                    "The number of return units "
                    "should match the number of returned values. "
                    f"{value=} {dimensions=} {self.definition.fortran_type=}"
                )

    @property
    def requires_units(self) -> bool:
        """
        Whether this value requires units or not

        Returns
        -------
            ``True`` if this value requires units, ``False`` otherwise.
        """
        return (not self.dynamic_unit) and self.definition.is_numeric_type



"""
Base utilities for the CLI
"""
from __future__ import annotations

import sys
from typing import Any

import click
from loguru import logger

DEFAULT_LOGGING_CONFIG = dict(
    handlers=[
        dict(
            sink=sys.stderr,
            colorize=True,
            format=" - ".join(
                [
                    "<green>{time:!UTC}</>",
                    "<lvl>{level}</>",
                    "<cyan>{name}:{file}:{line}</>",
                    "<lvl>{message}</>",
                ]
            ),
        )
    ],
)
"""Default configuration used with :meth:`loguru.logger.configure`"""


def setup_logging(config: dict[str, Any] | None = None) -> None:
    """
    Early setup for logging.

    Parameters
    ----------
    config
        Passed to :meth:`loguru.logger.configure`. If not passed,
        :const:`DEFAULT_LOGGING_CONFIG` is used.
    """
    if config is None:
        config = DEFAULT_LOGGING_CONFIG

    logger.configure(**config)
    logger.enable("fgen")


@click.group
def cli() -> None:
    """
    Entrypoint for the CLI
    """
    setup_logging()



"""
Generation of the Python wrapper module
"""
from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from typing import Optional, Union

from attrs import define

from fgen.data_models import (
    Method,
    Module,
    MultiReturn,
    Package,
    PackageSharedElements,
    Value,
)
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
    post_process_jinja_rendering,
)
from fgen.wrapper_building.formatting import format_python_code
from fgen.wrapping_strategies import (
    PassingToFortranSteps,
    WrappingStrategyLike,
    get_wrapping_strategy,
)


@define
class PythonWrapperModuleBuilder:
    """
    Builder of a Python wrapper module
    """

    package: Package
    """
    Package for which the builder is building wrappers
    """

    module: Module
    """
    Module for which to build the wrapper
    """

    shared: PackageSharedElements
    """
    Elements which have to be shared across the package

    For example, the names of functions which are used in more than one wrapper module.
    """

    @property
    def requires_union(self) -> bool:
        """
        Whether the module being wrapped requires ``typing.Union``
        """

        def needs_union(val: Union[MultiReturn, Value]) -> bool:
            fdt = val.definition.as_fortran_data_type()
            return fdt.is_derived_type and not fdt.is_pointer

        attributes_require_union = any(
            needs_union(a) for a in self.module.provides.exposed_attributes.values()
        )

        methods_require_union = any(
            any(needs_union(p) for p in method.parameters.values())
            for method in self.module.provides.methods.values()
        )

        return attributes_require_union or methods_require_union

    @property
    def requires_np(self) -> bool:
        """
        Whether the module being wrapped requires ``numpy as np``
        """

        def needs_np(val: Union[MultiReturn, Value]) -> bool:
            fdt = val.definition.as_fortran_data_type()
            return fdt.is_deferred_array

        attributes_require_np = any(
            needs_np(a) for a in self.module.provides.exposed_attributes.values()
        )

        methods_require_np = any(
            needs_np(method.returns)
            or any(needs_np(p) for p in method.parameters.values())
            for method in self.module.provides.methods.values()
        )

        return attributes_require_np or methods_require_np

    @property
    def requires_npt(self) -> bool:
        """
        Whether the module being wrapped requires ``numpy.typing as npt``
        """

        def needs_npt(val: Union[MultiReturn, Value]) -> bool:
            fdt = val.definition.as_fortran_data_type()
            return fdt.is_deferred_array and not val.dynamic_unit

        attributes_require_npt = any(
            needs_npt(a) for a in self.module.provides.exposed_attributes.values()
        )

        methods_require_npt = any(
            needs_npt(method.returns)
            or any(needs_npt(p) for p in method.parameters.values())
            for method in self.module.provides.methods.values()
        )

        return attributes_require_npt or methods_require_npt

    @property
    def requires_quantity(self) -> bool:
        """
        Whether the module being wrapped requires ``pint``'s quantity to be exposed
        """

        def needs_quantity(val: Union[MultiReturn, Value]) -> bool:
            return bool(val.dynamic_unit)

        attributes_require_quantity = any(
            needs_quantity(a) for a in self.module.provides.exposed_attributes.values()
        )

        methods_require_quantity = any(
            needs_quantity(method.returns)
            or any(needs_quantity(p) for p in method.parameters.values())
            for method in self.module.provides.methods.values()
        )

        return attributes_require_quantity or methods_require_quantity

    def get_wrapping_strategy(
        self, value: Union[MultiReturn, Value]
    ) -> WrappingStrategyLike:
        """
        Get wrapping strategy

        Parameters
        ----------
        value
            Value for which to get the wrapping strategy

        Returns
        -------
            Wrapping strategy to use with the value
        """
        return get_wrapping_strategy(value.definition.as_fortran_data_type())

    def get_module_level_docstring(self) -> str:
        """
        Get the module-level docstring

        Returns
        -------
            Module-level docstring
        """
        template = get_template_in_directory(
            "module-level-docstring.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            wrapper_module_name=self.module.wrapper_module_name,
            fortran_module_name=self.module.name,
        )

        return result

    def get_module_imports(self, extension: str) -> str:
        """
        Get the module's imports

        Parameters
        ----------
        extension
            The name of the overall extension module being built

        Returns
        -------
            Module ``import`` statements
        """
        template = get_template_in_directory(
            "module-imports.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(builder=self, extension=extension)

        return result

    def get_requirements_python_import_statements(self) -> str:
        """
        Get the Python ``import`` statements needed for the module's requirements

        These requirements are the requirements defined in
        ``self.module.requirements``.

        Returns
        -------
            Python ``import`` statements needed for the module
        """
        requirements_template = get_template_in_directory(
            "module-imports-requirements.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        required_imports = self.get_requirements_python_imports(self.module)

        out: list[str] = []
        for providing_module in sorted(required_imports.keys()):
            imports = sorted(required_imports[providing_module])
            imports_str = ", ".join(imports)

            rendered = requirements_template.render(
                providing_module=providing_module, imports_str=imports_str
            )
            out.append(rendered)

        return "\n".join(out)

    def get_enum_python_import_statements(self) -> Optional[str]:
        """
        Get the Python ``import`` statements needed for enums used by ``self.module``

        Returns
        -------
            Python ``import`` statements needed for any enums used by ``self.module``.
            If no imports are required for enums, ``None`` is returned.
        """
        required_enums = []
        to_check = [
            self.module.provides.attributes.values(),
            *[
                method.parameters.values()
                for method in self.module.provides.methods.values()
            ],
            *[[method.returns] for method in self.module.provides.methods.values()],
        ]

        for to_check_group in to_check:
            for to_check_value in to_check_group:
                to_check_fdt = to_check_value.definition.as_fortran_data_type()
                if to_check_fdt.is_enum and to_check_value not in required_enums:
                    required_enums.append(to_check_value)

        if not required_enums:
            return None

        providers = defaultdict(list)
        out: list[str] = []
        for required_enum in required_enums:
            fdt = required_enum.definition.as_fortran_data_type()
            python_equivalent_type = fdt.python_equivalent_type_annotation
            providing_requirement = self.module.get_requirement_that_provides(
                python_equivalent_type
            )
            providers[providing_requirement.python_module].append(
                python_equivalent_type
            )

        for provider, provides in providers.items():
            out.append(f"from {provider} import {','.join(set(provides))}")

        return "\n".join(out)

    def get_requirements_python_imports(self, module: Module) -> dict[str, list[str]]:
        """
        Get the Python imports needed for the module's requirements

        These requirements are the requirements defined in
        ``self.module.requirements``.

        Parameters
        ----------
        module
            Module for which to get the ``import`` statements

        Returns
        -------
            Modules (keys) and the types to import from them (values)
        """
        required_imports: dict[str, list[str]] = defaultdict(list)

        provided_type_attributes = module.provides.exposed_attributes.values()
        provided_type_method_arguments = [
            argument
            for method in module.provides.methods.values()
            for argument in method.parameters.values()
        ]
        provided_type_method_return_values = [
            method.returns for method in module.provides.methods.values()
        ]
        for value in [
            *provided_type_attributes,
            *provided_type_method_arguments,
            *provided_type_method_return_values,
        ]:
            if self.requires_import(value, module):
                base_type, needed_types = self.get_python_base_type_and_needed_types(
                    value
                )
                providing_requirement = module.get_requirement_that_provides(base_type)

                for needed_type in needed_types:
                    if (
                        needed_type
                        not in required_imports[providing_requirement.python_module]
                    ):
                        required_imports[providing_requirement.python_module].append(
                            needed_type
                        )

        return required_imports

    def requires_import(self, value: Union[Value, MultiReturn], module: Module) -> bool:
        """
        Return whether a value requires an import or not

        Parameters
        ----------
        value
            Value to check

        module
            Module in which the value is being used

        Returns
        -------
            ``True`` if this values requires an import, ``False`` otherwise.
        """
        fdt = value.definition.as_fortran_data_type()
        if not fdt.is_derived_type:
            return False

        # If the type we need is defined by the module, we don't need to import it
        return not (fdt.equivalent_python_type == module.provides.name)

    @staticmethod
    def get_python_base_type_and_needed_types(
        value: Union[Value, MultiReturn],
    ) -> tuple[str, tuple[str, ...]]:
        """
        Get the base type and the type we need to import

        Parameters
        ----------
        value
            Value for which to get the need type to import

        Returns
        -------
            Tuple. First element is the base type.
            The second type is the types that need to be imported.
        """
        base_type = value.definition.python_type_as_str()
        no_setters_type = f"{base_type}NoSetters"

        if value.definition.as_fortran_data_type().is_pointer:
            return (base_type, (base_type,))

        return (base_type, (base_type, no_setters_type))

    def get_module_units(self) -> str:
        """
        Get the module's units declaration

        Returns
        -------
            Module's units declaration
        """
        template = get_template_in_directory(
            "module-units.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        units = {}
        units_multi_return = {}
        for k, v in self.module.provides.units.items():
            if isinstance(v, str):
                units[k] = v
            elif isinstance(v, tuple):
                units_multi_return[k] = v
            else:  # pragma: no cover
                # Should be impossible to get here, but just in case
                msg = f"We don't yet support units of type {type(v)} here"  # type: ignore[unreachable]
                raise NotImplementedError(msg)

        result = template.render(
            units=units,
            units_multi_return=units_multi_return,
        )

        return result

    def get_provided_type_docstring(self) -> str:
        """
        Get the provided type's docstring

        Returns
        -------
            Provided type's docstring
        """
        template = get_template_in_directory(
            "provided-type-docstring.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            provided_type=self.module.provides.name,
            provided_type_description=self.module.provides.description,
        )

        return result

    def get_verify_units_decorator(
        self,
        arguments: Iterable[Union[MultiReturn, Value]],
        return_value: Optional[Union[MultiReturn, Value]] = None,
        include_self_argument: bool = True,
    ) -> str:
        """
        Get the ``verify_units`` decorator for a callable

        Parameters
        ----------
        arguments
            Arguments for the callable

        return_value
            Return value from the callable

        include_self_argument
            Should we assume that there is a ``self`` argument to the callable too?

        Returns
        -------
            ``verify_units`` decorator for the callable
        """
        template = get_template_in_directory(
            "verify-units-decorator.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        arguments_info_for_verify_units: list[Union[str, None]] = []
        if include_self_argument:
            arguments_info_for_verify_units.append(None)

        for arg in arguments:
            if not self.include_in_python_callable_arguments(arg):
                continue

            if arg.unit:
                arguments_info_for_verify_units.append(
                    f'_UNITS["{self.get_python_user_facing_name(arg)}"]'
                )
            else:
                arguments_info_for_verify_units.append(None)

        return_info_for_verify_units: Union[None, str, list[str]] = None
        if (return_value is not None) and return_value.requires_units:
            if return_value.unit is None:  # pragma: no cover
                msg = (
                    "How did we get here? "
                    "Should have failed at initialisation of {return_value}"
                )
                raise AssertionError(msg)

            python_facing_name = self.get_python_user_facing_name(return_value)
            if isinstance(return_value, MultiReturn):
                return_info_for_verify_units = (
                    f'_UNITS_MULTI_RETURN["{python_facing_name}"]'
                )
            else:
                return_info_for_verify_units = f'_UNITS["{python_facing_name}"]'

        result = template.render(
            arguments_info_for_verify_units=arguments_info_for_verify_units,
            return_info_for_verify_units=return_info_for_verify_units,
        )

        return result

    def get_post_verify_units_argument_list(
        self, arguments: Iterable[Union[MultiReturn, Value]]
    ) -> str:
        r"""
        Get argument list to be used, assuming the ``verify_units`` decorator is applied

        The argument list includes the type hints that are expected.
        These type hints are created on the assumption that the ``verify_units``
        is being used on the callable.

        Parameters
        ----------
        arguments
            Arguments for which to generate the arguments list.

        Returns
        -------
            Arguments list, in the form "argument: type,\narg: type_arg\n..."
        """
        out = []
        for arg in arguments:
            if not self.include_in_python_callable_arguments(arg):
                continue

            ws = self.get_wrapping_strategy(arg)
            name = ws.get_python_user_facing_name(arg)
            type_hint = ws.get_python_post_verify_units_input_type_annotation(arg)
            out.append(f"{name}: {type_hint},")

        return "\n".join(out)

    def get_provided_type_repr_methods(self) -> str:
        """
        Get the provided type's representation relevant methods

        Returns
        -------
            Representation-relevant methods for the provided type
        """
        template = get_template_in_directory(
            "provided-type-str-repr-methods.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            exposed_attributes=self.module.provides.exposed_attributes,
        )

        return result

    def get_provided_type_class_methods(self, class_name: str) -> str:
        """
        Get the provided type's class methods

        Parameters
        ----------
        class_name
            Name of the class for which we are getting the class methods.

        Returns
        -------
            Class methods
        """
        template = get_template_in_directory(
            "provided-type-class-methods.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            builder=self,
            shared=self.shared,
            class_name=class_name,
        )

        return result

    def get_provided_type_getters_and_setters(
        self, include_setters: bool = True
    ) -> str:
        """
        Get the provided type's getters and setters

        Parameters
        ----------
        include_setters
            Should the setters be created too?

        Returns
        -------
            Provided type's getters and setters
        """
        template = get_template_in_directory(
            "getters-and-setters.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            builder=self,
            attributes=self.module.provides.exposed_attributes.values(),
            include_setters=include_setters,
        )

        return result

    def get_provided_type_methods(self) -> str:
        """
        Get the provided type's methods

        Returns
        -------
            Provided type's methods
        """
        template = get_template_in_directory(
            "methods.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            builder=self, methods=self.module.provides.methods.values()
        )

        return result

    def get_fortran_wrapper_method(self, method: Method) -> str:
        """
        Get the name of the method in the Fotran wrapper

        Parameters
        ----------
        method
            Method for which to get the name

        Returns
        -------
            Name of the method in the Fortran wrapper
        """
        return f"i_{method.name}"

    def get_passing_to_fortran_steps(
        self,
        arguments: Iterable[Union[MultiReturn, Value]],
        dynamic_unit: Optional[str] = None,
    ) -> PassingToFortranSteps:
        """
        Get passing to Fortran steps for a set of arguments

        Parameters
        ----------
        arguments
            Arguments for the callable. We get the Fortran passing steps for these.

        dynamic_unit
            If provided, specifies the source of the units
            to which we want to convert the arguments.

        Returns
        -------
            Steps for passing the arguments to Fortran
        """
        preparatory_python_calls_list = []
        fortran_module_callable_args: list[tuple[str, str]] = []
        for argument in arguments:
            if not self.include_in_python_callable_arguments(argument):
                # Let wrappers handle this
                continue

            ws = self.get_wrapping_strategy(argument)
            if argument.dynamic_unit:
                if isinstance(argument.dynamic_unit, str):
                    dynamic_unit_argument = argument.dynamic_unit

                elif dynamic_unit is None:
                    msg = "If dynamic unit is True, the source must be provided"
                    raise AssertionError(msg)

                else:
                    dynamic_unit_argument = dynamic_unit

            else:
                dynamic_unit_argument = None

            argument_steps = ws.get_passing_to_fortran_steps(
                value=argument, dynamic_unit=dynamic_unit_argument
            )

            if argument_steps.preparatory_python_calls is not None:
                preparatory_python_calls_list.append(
                    argument_steps.preparatory_python_calls
                )

            fortran_module_callable_args.extend(
                argument_steps.fortran_module_callable_args
            )

        if not preparatory_python_calls_list:
            preparatory_python_calls = None
        else:
            preparatory_python_calls = "\n".join(preparatory_python_calls_list)

        return PassingToFortranSteps(
            preparatory_python_calls=preparatory_python_calls,
            fortran_module_callable_args=tuple(fortran_module_callable_args),
        )

    def get_python_user_facing_name(self, value: Union[MultiReturn, Value]) -> str:
        """
        Get the Python-user facing name of a value

        Parameters
        ----------
        value
            Value for which to get the Python user-facing name

        Returns
        -------
            Python user-facing name
        """
        ws = self.get_wrapping_strategy(value)
        return ws.get_python_user_facing_name(value)

    def include_in_python_callable_arguments(
        self, value: Union[MultiReturn, Value]
    ) -> bool:
        """
        Determine whether a value should be included as an argument to Python callables

        Parameters
        ----------
        value
            Value to check

        Returns
        -------
            ``True`` if the value should be included as an argument to Python callables,
            ``False`` otherwise.
        """
        return not value.definition.is_fortran_units_holder


def generate_python_wrapper_module(
    builder: PythonWrapperModuleBuilder, extension: str
) -> str:
    """
    Generate the Python wrapper module

    Parameters
    ----------
    builder
        Builder to use to generate the Python wrapper module

    extension
        Name of the extension module that will contain the compiled wrappers

    Returns
    -------
        Python wrapper module as code
    """
    template = get_template_in_directory(
        "python-wrapper-module.py.jinja", Path(__file__).parent, JINJA_ENV
    )

    result = post_process_jinja_rendering(
        template.render(
            builder=builder,
            extension=extension,
        )
    )

    return format_python_code(result)



"""
Package data model
"""

from __future__ import annotations

from collections import defaultdict
from typing import Any, Optional, Union

import attr
from attrs import define, field

from fgen.data_models.module import Module
from fgen.data_models.module_enum_defining import ModuleEnumDefining
from fgen.data_models.multi_return import MultiReturn
from fgen.data_models.value import Value


class NoProvidingModuleError(ValueError):
    """
    Exception raised when no module provides a given type
    """

    def __init__(
        self, sought_type: str, value: Union[MultiReturn, Value], package: Package
    ):
        error_msg = (
            f"No module provides {sought_type}, which is the type of {value}. "
            f"We searched in {package=}."
        )

        super().__init__(error_msg)


@define
class Package:
    """
    Data model of a package

    It isn't clear that our package is exactly like a typical Python/Fortran package,
    so it is best to think of it exactly as it is for now:
    a Package is a collection of :obj:~`fgen.data_models.Module`'s.
    While the naming is similar to the Package > Module hierarchy used by Python,
    they should be treated as different concepts.
    """

    modules: tuple[Module, ...] = field()
    """
    Collection of modules that define derived types used within the package
    """

    modules_enum_defining: tuple[ModuleEnumDefining, ...] = field(factory=tuple)
    """
    Collection of modules that define enums used within the package
    """

    @modules.validator
    def _modules_provide_unique(
        self, attribute: attr.Attribute[Any], value: tuple[Module, ...]
    ) -> None:
        provided_derived_types_modules = defaultdict(list)

        # Multiple loops as mypy being silly
        for module in value:
            provided_derived_types_modules[module.provides.name].append(module.name)

        for module_enum_defining in self.modules_enum_defining:
            provided_derived_types_modules[module_enum_defining.provides.name].append(
                module_enum_defining.name
            )

        not_unique = tuple(
            (k, sorted(v))
            for k, v in provided_derived_types_modules.items()
            if len(v) > 1
        )
        if not_unique:
            duplicate_type_info = "\n".join(
                (
                    f"Provided type `{provided_type}` is provided by: {provided_by}"
                    for provided_type, provided_by in not_unique
                )
            )
            msg = (
                "The following derived types are provided by more than one module: "
                f"{duplicate_type_info}"
            )

            raise ValueError(msg)

    # Can cache this if we need speed
    def get_module_that_provides_values_type(
        self, value: Union[MultiReturn, Value]
    ) -> Module:
        """
        Get the module that provides a value's data type

        Parameters
        ----------
        value
            Value of which to find the data type provider

        Returns
        -------
            Module that provides ``value``

        Raises
        ------
        NoProvidingModuleError
            No module that provides ``value``'s data type is part of ``self.modules``.
        """
        sought_type = value.definition.as_fortran_data_type().equivalent_python_type
        for module in self.modules:
            if module.provides.name == sought_type:
                return module

        raise NoProvidingModuleError(sought_type=sought_type, value=value, package=self)

    def find_providing_module(
        self, value: Union[MultiReturn, Value]
    ) -> Optional[Module]:
        """
        Find the module that provides a value's data type

        If no module provides the value's data type, ``None`` is returned.

        Parameters
        ----------
        value
            Value of which to find the data type provider

        Returns
        -------
            Module that provides ``value``.
            If no module provides the value, ``None`` is returned.
        """
        try:
            return self.get_module_that_provides_values_type(value)
        except NoProvidingModuleError:
            return None



"""
Shared elements of a package data model
"""
from __future__ import annotations

from attrs import define


@define
class PackageSharedElements:
    """
    Data model of the shared elements in a package

    This helps us define the shared elements across the package.
    For example, the names of callables
    that are defined in Fortran but used in Python.
    """

    @property
    def from_build_args_name(self) -> str:
        """
        Name of the Python function which creates an object from its build arguments
        """
        return "from_build_args"

    @property
    def finalise_method_name(self) -> str:
        """
        Name of the callables which finalise instances
        """
        return "finalize"

    @property
    def fortran_build_callable_name(self) -> str:
        """
        Name of Fortran callables which build instances
        """
        return "instance_build"

    @property
    def fortran_get_free_instance_number_callable_name(self) -> str:
        """
        Name of Fortran callables which get the number of a free/available instance
        """
        return "get_free_instance_number"

    @property
    def fortran_get_instance_callable_name(self) -> str:
        """
        Name of Fortran callables which get an instance
        """
        return "get_instance"

    @property
    def fortran_instance_finalise_callable_name(self) -> str:
        """
        Name of Fortran callables finalise an instance
        """
        return "instance_finalize"

    @property
    def fortran_wrapper_method_prefix(self) -> str:
        """
        Prefix to use when creating the Fortran wrapper equivalent of methods
        """
        return "i_"  # i for 'instance'

    @property
    def n_fortran_instances_to_expose(self) -> int:
        """
        Number of Fortran instances to expose

        Once the user requests more instances than this, an error is raised.
        """
        return 4096

    @property
    def no_setters_suffix(self) -> str:
        """
        Suffix to use when creating the no setters version of our wrapper classes
        """
        return "NoSetters"



"""
:mod:`fgen`'s wrapper building module
"""

from __future__ import annotations

import functools
from pathlib import Path
from typing import Callable, Protocol, TypeVar, Union

from loguru import logger

from fgen.data_models import Module, Package, PackageSharedElements
from fgen.wrapper_building.fortran_manager_module import (
    FortranManagerModuleBuilder,
    generate_fortran_manager_module,
)
from fgen.wrapper_building.fortran_wrapper_module import (
    FortranWrapperModuleBuilder,
    generate_fortran_wrapper_module,
)
from fgen.wrapper_building.python_enums_module import generate_python_enums_module
from fgen.wrapper_building.python_init_module import generate_python_init_module
from fgen.wrapper_building.python_wrapper_module import (
    PythonWrapperModuleBuilder,
    generate_python_wrapper_module,
)

AvailableBuilders = Union[
    FortranManagerModuleBuilder, FortranWrapperModuleBuilder, PythonWrapperModuleBuilder
]
T_builder = TypeVar("T_builder", bound=AvailableBuilders)
T_contra = TypeVar("T_contra", contravariant=True)


def try_to_write_file(file: Path, contents: str) -> None:
    """
    Try to write a file to disk, logging an exception if this fails.

    Parameters
    ----------
    file
        File to write to.

    contents
        Contents to write in the file.
    """
    try:
        with open(file, mode="w", encoding="utf-8") as fh:
            fh.write(contents)
    except Exception:
        logger.exception(
            f"Error encountered while trying to write {file=} with {contents=}"
        )
        raise


def write_file_with_checks(
    file: Path,
    desc: str,
    generate_content: Callable[[], str],
    keep_existing: bool = False,
    force: bool = False,
) -> None:
    """
    Write a file, first checking what to do if the file already exists.

    Parameters
    ----------
    file
        Path in which the file will be written if the checks pass.

    desc
        Description of the file which will be written if the checks pass.
        Only used for logging.

    generate_content
        Function which generates the contents of the file.
        This is only called if the file is to be written.

    keep_existing
        Whether to always keep existing files.
        If this is ``True``, if a file exists, it is not overwritten.

    force
        Whether to force overwrite an existing file.
        If this is ``True``, if a file exists, it is overwritten.

    Raises
    ------
    FileExistsError
        ``file`` exists and no flags are set to specify what to do.

    ValueError
        Both ``keep_existing`` and ``force`` are ``True``.
        ``force`` will never be used in this case, hence an error is raised.
    """
    if keep_existing and force:
        msg = (
            "Both ``keep_existing`` and ``force`` are ``True``. "
            "``force`` will never be used in this case. "
            "Please update your arguments."
        )
        raise ValueError(msg)

    if file.exists():
        if keep_existing:
            logger.info(f"Existing file will be kept: {file=}")
            return

        if not force:
            logger.error("Existing file should not be overwritten")
            raise FileExistsError(file)

        logger.warning(
            f"{force=}, existing file will be forcefully overwritten: {file=}"
        )

    contents = generate_content()
    try_to_write_file(file, contents)
    logger.info(f"Wrote {desc} to {file}")


class ModuleWrapperGeneratorCallableLike(Protocol[T_contra]):
    """
    Callable that can generate a wrapper for a module
    """

    def __call__(self, builder: T_contra) -> str:
        """
        Generate the wrapper
        """


def write_module_wrappers(  # noqa: PLR0913
    package: Package,
    shared: PackageSharedElements,
    module: Module,
    manager_directory: Path,
    wrapper_directory: Path,
    python_directory: Path,
    extension: str,
    force: bool,
) -> None:
    """
    Write the wrappers for a module

    Parameters
    ----------
    package
        Package this module belongs to

    shared
        Shared elements across ``package``

    module
        Module to write wrappers for

    manager_directory
        Directory in which to write the manager wrappers

    wrapper_directory
        Directory in which to write the Fortran wrappers

    python_directory
        Directory in which to write the Python wrappers

    extension
        Name of the extension module that will be called from Python

    force
        Whether to force overwrite existing files or not
    """
    wmw_partial = functools.partial(
        write_module_wrapper,
        force=force,
    )
    wmw_partial(
        desc="Fortran manager module",
        generator_function=generate_fortran_manager_module,
        builder=FortranManagerModuleBuilder(
            package=package, shared=shared, module=module
        ),
        file_suffix="_manager.f90",
        outdir=manager_directory,
    )
    wmw_partial(
        desc="Fortran wrapper module",
        generator_function=generate_fortran_wrapper_module,
        builder=FortranWrapperModuleBuilder(
            package=package, shared=shared, module=module
        ),
        file_suffix="_wrapped.f90",
        outdir=wrapper_directory,
    )
    wmw_partial(
        desc="Python wrapper module",
        generator_function=functools.partial(
            generate_python_wrapper_module, extension=extension
        ),
        builder=PythonWrapperModuleBuilder(
            package=package, shared=shared, module=module
        ),
        file_suffix=".py",
        outdir=python_directory,
    )


def write_module_wrapper(  # noqa: PLR0913
    desc: str,
    generator_function: ModuleWrapperGeneratorCallableLike[T_builder],
    builder: T_builder,
    file_suffix: str,
    outdir: Path,
    force: bool,
) -> None:
    """
    Write a wrapper for a module

    Parameters
    ----------
    desc
        Description of the wrapper being written (used for logging)

    generator_function
        Function that generates the wrapper's contents

    builder
        The builder to use in combination with ``generator_function``
        to generate the wrapper's contents

    file_suffix
        The suffix to apply to the builder's truncated name.
        This combination defines the filename.

    outdir
        Directory in which to write the wrapper.

    force
        Whether to force overwrite existing files or not.
    """
    write_file_with_checks(
        file=outdir / f"{builder.module.truncated_name}{file_suffix}",
        desc=desc,
        generate_content=functools.partial(generator_function, builder=builder),
        force=force,
    )


class PackageWrapperGeneratorCallableLike(Protocol):
    """
    Callable that can generate a wrapper for a package
    """

    def __call__(self, package: Package) -> str:
        """
        Generate the wrapper
        """


def write_package_wrapper(  # noqa: PLR0913
    package: Package,
    desc: str,
    generator_function: PackageWrapperGeneratorCallableLike,
    outdir: Path,
    filename: str,
    keep_existing: bool = False,
    force: bool = False,
) -> None:
    """
    Write a wrapper for the package

    Parameters
    ----------
    package
        Package for which to write the wrapper.

    desc
        Description of the wrapper being written (used for logging).

    generator_function
        Function to call to generate the wrapper's contents.

    outdir
        Directory in which to write the wrapper.

    filename
        Name of the file in which to write the wrapper.

    keep_existing
        Whether to keep existing files, irrespective of the value of ``force``.

    force
        Whether to force overwrite existing files.
    """
    write_file_with_checks(
        file=outdir / filename,
        desc=desc,
        generate_content=functools.partial(generator_function, package=package),
        force=force,
        keep_existing=keep_existing,
    )


def process_package(  # noqa: PLR0913
    package: Package,
    shared: PackageSharedElements,
    extension: str,
    manager_directory: Path,
    wrapper_directory: Path,
    python_directory: Path,
    force: bool = True,
) -> None:
    """
    Auto-generate python and fortran code for a package

    This function writes 3 files for each module that is being wrapped:

    - manager: Lifecycle handling for derived types (in Fortran)

    - wrapper: Provides a fortran wrapper to the generated functions that can be
                made available to Python

    - python module: Python interface for creating and manipulating derived types.

    This function may also write other files necessary
    for wrapping the package as a whole.
    This includes generating a `__init__.py` file
    if it doesn't already exist in the python directory.

    Parameters
    ----------
    package
        Package to generate wrappers for

    shared
        Shared elements which should be consistent
        across the package's generated wrappers

    extension
        Package name for the compiled extension module

        Typically, this is of the form ``my_module._lib``

    manager_directory
        Directory in which to write the derived type lifecycle manager(s)

    wrapper_directory
        Directory to write the Python-Fortran wrapper(s).

    python_directory
        Directory in which to write the python module(s).

    force
        If ``True``, overwrite any existing files

    Raises
    ------
    FileExistsError
        If ``not force`` and a targeted file exists
    """
    write_package_wrapper(
        package=package,
        desc="Python package `__init__.py` file",
        generator_function=generate_python_init_module,
        outdir=python_directory,
        filename="__init__.py",
        keep_existing=True,
    )
    if package.modules_enum_defining:
        write_package_wrapper(
            package=package,
            desc="Python equivalent of enums",
            generator_function=generate_python_enums_module,
            outdir=python_directory,
            filename="enums.py",  # Can make this configurable too if we want
            # # Not sure what the right choice for keep_existing is.
            # keep_existing=True,
        )

    for module in package.modules:
        write_module_wrappers(
            package=package,
            shared=shared,
            module=module,
            manager_directory=manager_directory,
            wrapper_directory=wrapper_directory,
            python_directory=python_directory,
            extension=extension,
            force=force,
        )


__all__ = [
    "generate_fortran_manager_module",
    "generate_fortran_wrapper_module",
    "generate_python_init_module",
    "generate_python_wrapper_module",
    "process_package",
]



"""
Support for parsing Fortran type declarations

We just want to make this as helpful as possible
when auto-generating our wrappers
while keeping the implemenation here as minimal as possible
because duplicating the Fortran implementation logic is not the point of this module.
The fortran compiler will catch any odd type declarations at compile-time.

To understand what is going on,
we start with the syntax for declaring fortran types as defined in
Section 5.1 of the `Fortran 2003 standard <https://j3-fortran.org/doc/year/04/04-007.pdf>`_.
It looks like the below:

.. syntax higlighting below is wrong, not sure what language the specs are given in
.. code-block:: fortran

    R501 type-declaration-stmt  is declaration-type-spec [ [ , attr-spec ] ... :: ] entity-decl-list
    R502 declaration-type-spec  is intrinsic-type-spec
                                or TYPE ( derived-type-spec )
                                or CLASS ( derived-type-spec )
                                or CLASS ( * )

In short, a type declaration statement is composed of a declaration type specification,
a collection of one or more attribute specifications
and then an entity declaration list.
The declaration type specification can be either an intrinsic type
that has been defined by the language,
or a derived type which has been declared using a ``type`` block elsewhere.

We focus on the combination of the declaration type specification
(which we typically shorten to type specification)
and the attribute specifications i.e. we largely ignore the entity declarations.
In the absence of a defined term in the Fortran standard,
we call this combination the type attribute declaration throughout.

To illustrate, consider the following example

.. code-block:: fortran

    real(8), dimension(3), pointer :: heat_uptake

Here is how we would refer to the parts of this example:

- type declaration statement: ``real(8), dimension(3), pointer :: heat_uptake``
- declaration type specification (or simply type specification): ``real(8)``
- attribute specifications: ``dimension(3), pointer``
- type attribute declaration: ``real(8), dimension(3), pointer``
- entity declaration (largely ignored by this module, we treat the double colon ``::`` as part of the entity declaration): ``:: heat_uptake``

One other example using a derived type

.. code-block:: fortran

    type(my_calculator) :: heat_uptake_calculator

Here is how we would refer to the parts of this example:

- type declaration statement: ``type(my_calculator) :: heat_uptake_calculator``
- declaration type specification (or simply type specification): ``type(my_calculator)``
- attribute specifications: none
- type attribute declaration (same as type specification in this case): ``type(my_calculator)``
- entity declaration (largely ignored by this module, we treat the double colon ``::`` as part of the entity declaration): ``:: heat_uptake_calculator``

Notes
-----
Fortran type declaration statements (and Fortran in general) are case-insensitive

This module relies heavily on regular expressions. If needed, use an online tool like
`regex101.com <https://regex101.com/>`_ to help you understand them.
"""  # noqa: E501
from __future__ import annotations

import re
from collections.abc import Sequence
from typing import Any, Optional, Union

from attrs import define, field

_AT_LEAST_ONE_ALPHANUMERIC_REGEXP: str = r"[a-z0-9]+"
"""Regex representing a search for at least one alphanumeric character"""

_DIMENSION_SPECIFICATION: str = f"(\\*|:|{_AT_LEAST_ONE_ALPHANUMERIC_REGEXP})"
"""
Regex representing a possible dimension attribute specification in a Fortran definition

This also captures deferred and assumed shaped arrays
which are not currently supported by fgen,
but may be supported in future.

e.g. "(4, 3)", "(n, m)", "(5)", "(n, 4)"
"""

_DIMENSION_REGEX: str = (
    r"DIMENSION\("
    r"(?P<dimension>"
    f"{_DIMENSION_SPECIFICATION}(,\\s*{_DIMENSION_SPECIFICATION})*"
    ")"
    r"\)"
)
"""
Regex string used to check for dimension information within attribute specifications

e.g. "dimension(4, 3)", "dimension(n)"
"""

_KIND_SPECIFICATION: str = r"(\(.*\)|)$"
"""
Regex string to capture optional type modifiers

e.g. "(kind=8)" in "real(kind=8)"
"""

_INTRINSIC_TYPE_SPECIFICATIONS: tuple[str, ...] = (
    "INTEGER",
    "REAL",
    "COMPLEX",
    "CHARACTER",
    "LOGICAL",
)
"""Intrinsic Fortran type specifications that we support (case insensitive)"""

_ENUM_LIKE_REGEX: str = r"integer\(\s?kind\s?\(\s?(?P<enum_name>[a-z0-9]+)\s?\)\s?\)"
"""
Regex string to capture Fortran enum-like specifications

Matches things like "integer(kind(Enum))" and "integer(kind(EnumValue))".
The enum type is captured in a group called "enum_name".
"""

_DERIVED_TYPE_REGEX: str = r"type\(\s?(?P<derived_type_name>[a-z0-9]+)\s?\)"
"""
Regex string to capture derived type specifications

Matches things like "type(my_derived_type)" and "type(calculator)".
The derived type is captured in a group called "derived_type_name".
We currently don't support "class(abc)" specifications.
"""

SUPPORTED_TYPE_SPECIFICATIONS: tuple[str, ...] = (
    *[
        f"^{intrinsic_type_spec}{_KIND_SPECIFICATION}"
        for intrinsic_type_spec in _INTRINSIC_TYPE_SPECIFICATIONS
    ],
    _DERIVED_TYPE_REGEX,
    _ENUM_LIKE_REGEX,
)
"""
The collection of valid type specifications that are supported

This covers both intrinsic type specifications and derived type specifications.
"""

SUPPORTED_ATTRIBUTE_SPECIFICATIONS: tuple[Union[str, re.Pattern[str]], ...] = (
    "ALLOCATABLE",
    # "ASYNCHRONOUS",
    _DIMENSION_REGEX,
    # "INTENT",
    # "INTRINSIC",
    # "language-binding-spec"
    # "OPTIONAL",
    "POINTER",
    # "PROTECTED",
    # "SAVE",
    "TARGET",
    # "VALUE",
    # "VOLATILE",
)
"""
The collection of valid attribute specifications that are supported

In the type declaration, "type(my_calculator), pointer",
the attribute specification is "pointer"
(the other part, "type(my_calculator)",
is simply the type specification
(or declaration type specification if you're being very precise))

Note that only a subset of valid attributes can be specified.
The more esoteric attributes have been excluded
until the use case for these attributes is better understood.
"""


def _get_type_attribute_declaration(type_declaration_statement: str) -> str:
    """
    Get type attribute declaration from a type declaration statement

    Simply removes the entity declaration
    """
    return type_declaration_statement.split("::")[0]


def _get_parts(type_declaration_statement: str) -> tuple[str, tuple[str, ...]]:
    """
    Split a type declaration statement into its parts

    Specifically, into a type specification and a collection of attribute specifications

    This function is used below by :meth:`FortranDataType.as_str`.
    That method takes care of
    validating the different parts of a Fortran type declaration statement.

    See the docstring of :mod:`fgen.fortran_parsing`
    for further details on how we interpret Fortran type declaration statements.

    Parameters
    ----------
    type_declaration_statement
        Type declaration statement to parse

        The entity declaration component
        (any parts including and after the "::")
        is optional and will be ignored

    Returns
    -------
        The type specification and any additional attribute specifications

    Examples
    --------
    >>> _get_parts("real :: my_variable")
    ('real', ())
    >>> _get_parts("real, dimension(5)")
    ('real', ('dimension(5)',))
    >>> _get_parts("real, dimension(5, 3)")
    ('real', ('dimension(5, 3)',))
    >>> _get_parts("integer, dimension(:, :)")
    ('integer', ('dimension(:, :)',))
    >>> _get_parts("logical, dimension(2, :)")
    ('logical', ('dimension(2, :)',))
    >>> _get_parts("logical, dimension(2, :), pointer")
    ('logical', ('dimension(2, :)', 'pointer'))
    >>> _get_parts("integer(kind(EnumValue))")
    ('integer(kind(EnumValue))', ())
    """
    type_attribute_declaration = _get_type_attribute_declaration(
        type_declaration_statement
    )

    # Check if this is enum type, such things break everything below
    enum_regex_match = re.match(
        _ENUM_LIKE_REGEX, type_attribute_declaration, flags=re.IGNORECASE
    )
    if enum_regex_match:
        return type_attribute_declaration, tuple()

    part_regexp_including_brackets = r"[^,\s]+(?:\([^\(]*?\))"
    part_regexp_no_brackets = r"[^,\s]+"
    part_regexp: str = "|".join(
        [part_regexp_including_brackets, part_regexp_no_brackets]
    )
    """
    Regular expression that captures the parts of the type declaration statement

    It effectively just splits the string into comma separated pieces without
    surrounding whitespace (with a little extra piece to make sure you don't
    get nested brackets (e.g. "dimension((4, 3))") in any part, which is invalid
    Fortran as far as we know anyway)

    e.g. this can be used to split "real, dimension(2, 3, 4)"
    into ["real", "dimension(2, 3, 4)"]
    """
    parts = re.findall(part_regexp, type_attribute_declaration)

    if not len(parts):
        raise ValueError(  # noqa: TRY003
            "An invalid type attribute declaration was provided: "
            f"{type_attribute_declaration!r}"
        )

    # The Fortran spec states that the first part must be the type specification and
    # the rest are attribute specifications so we can safely assume order here
    return parts[0], tuple(parts[1:])


def _validate_fortran_type_attribute_declaration(
    type_specification: str, attribute_specifications: tuple[str, ...]
) -> None:
    """
    Validate that a fortran type attribute declaration is supported by fgen

    The type attribute declaration must have been split
    (using e.g. :func:`_get_parts`)
    before using this function.

    We don't support all valid fortran.
    See the examples below and the tests for examples of supported options.

    This validation is pretty crude as invalid Fortran will be caught at compile time.

    See the docstring of :mod:`fgen.fortran_parsing`
    for further details on how we interpret Fortran type declaration statements.

    Parameters
    ----------
    type_specification
        Type specification

    attribute_specifications
        Attribute specifications

        See :data:`SUPPORTED_ATTRIBUTE_SPECIFICATIONS`
        for the attributes that are currently supported by fgen

    Raises
    ------
    ValueError
        An unsupported fortran type specification is supplied

    Examples
    --------
    Below are some examples of (split) fortran type declarations that pass

    >>> _validate_fortran_type_attribute_declaration("integer", ())
    >>> _validate_fortran_type_attribute_declaration("real", ("dimension(5)",))
    >>> _validate_fortran_type_attribute_declaration(
    ...     "type(calculator)", ("dimension(5)", "pointer")
    ... )

    See Also
    --------
    :func:`_get_parts`
    """
    # Any attributes must match the regex of a supported attribute
    if not any(
        re.match(supported_type, type_specification, flags=re.IGNORECASE)
        for supported_type in SUPPORTED_TYPE_SPECIFICATIONS
    ):
        raise ValueError(  # noqa: TRY003
            f"Unsupported type specification: {type_specification}"
        )

    # Any attributes must match the regex of a supported attribute
    for attribute_specification in attribute_specifications:
        if not any(
            re.match(attribute_regex, attribute_specification, flags=re.IGNORECASE)
            for attribute_regex in SUPPORTED_ATTRIBUTE_SPECIFICATIONS
        ):
            raise ValueError(  # noqa: TRY003
                f"Unsupported attribute specification: {attribute_specification}"
            )


def _convert_complex_attribute_specifications(
    attribute_specifications: Sequence[str],
) -> tuple[Union[str, DimensionAttributeSpecification], ...]:
    """
    Convert any complex attribute specifications to their supporting classes

    For example, a dimension attribute specification is converted to a
    :class:`DimensionAttributeSpecification`
    which holds additional information about the attribute.
    If the attribute specification doesn't have a matching class
    then it will remain as a string.

    Parameters
    ----------
    attribute_specifications
        Collection of attribute specifications

    Returns
    -------
        Tuple that contains a combination of strings and classes in the case of the more
        complicated attributes
    """
    attributes: list[Union[str, DimensionAttributeSpecification]] = list(
        attribute_specifications
    )

    for i in range(len(attribute_specifications)):
        match = re.search(
            _DIMENSION_REGEX, attribute_specifications[i], flags=re.IGNORECASE
        )
        if match:
            attributes[i] = DimensionAttributeSpecification.from_dimension_info(
                match.group("dimension")
            )

    return tuple(attributes)


@define
class DimensionAttributeSpecification:
    """
    Dimension attribute specification

    Defines the shape of a data type. See Section 5.1.2.5 of the
    `Fortran 2003 standard <https://j3-fortran.org/doc/year/04/04-007.pdf>`_
    for a description of the dimension attribute specification.

    Currently, we only support explicit dimensions
    where the size of the dimension
    is explicitly declared at compile-time with an integer or execution-time
    using a variable provided to a procedure
    (e.g. using "n" which is a variable that is also provided to a procedure).
    The later form is an automatic explicit dimension
    and will rely upon the Fortran compiler to validate the attribute declaration
    since additional context is required about where the attribute is used.

    Assumed-sized dimensions ("*") are not currently supported.
    """

    dimensions: tuple[Union[str, int], ...] = field()
    """
    Collection of dimensions of the attribute specification

    Can include explicit dimensions (e.g. 5)
    or automatic explicit dimensions (e.g. "n").
    Deferred dimensions (e.g. ":") are allowed.
    """

    @dimensions.validator
    def _check_dimensions(
        self, attribute: Any, value: tuple[Union[str, int], ...]
    ) -> None:
        for dimension in value:
            if dimension == "*":
                raise ValueError(  # noqa: TRY003
                    "Assumed-sized dimensions are not supported"
                )

    @classmethod
    def from_dimension_info(
        cls, dimension_info: str
    ) -> DimensionAttributeSpecification:
        """
        Create a DimensionAttributeSpecification from the dimension information

        The dimension information describes the shape and dimensionality of a variable.

        Parameters
        ----------
        dimension_info
            The dimension information
            is the content between the () for a dimension attribute.
            For example, for an attribute "dimension(8, n)"
            the dimension_info would be `"8, n"`.

            Static dimensions are converted to integers before initialising.

        Returns
        -------
            New DimensionAttributeSpecification
        """
        toks = dimension_info.split(",")

        type_info: list[Union[str, int]] = []
        for tok in toks:
            try:
                size: int | str = int(tok)
            except ValueError:
                # assume : or n or something i.e. not static size
                size = tok.strip()

            type_info.append(size)

        return DimensionAttributeSpecification(dimensions=tuple(type_info))

    @property
    def ndim(self) -> int:
        """
        Number of dimensions of the attribute
        """
        return len(self.dimensions)

    def __str__(self) -> str:
        dimensions_as_str = ", ".join(str(dim) for dim in self.dimensions)
        return f"dimension({dimensions_as_str})"


def _get_python_type_with_dimensions(
    dimension_info: Sequence[Union[str, int]],
    base: str,
) -> str:
    """
    Get a Python type that supports dimension information

    See [#1](https://gitlab.com/magicc/fgen/-/issues/1)
    for ongoing discussions about updating this
    to return more sophisticated Python types.

    Parameters
    ----------
    dimension_info
        Dimension information
        (e.g. extracted elsewhere from a Fortran type declaration statement)

    base
        Base Python type to which the dimension information applies
        (e.g. "float", "int", "str")

    Returns
    -------
        Python type with dimension information

    Examples
    --------
    >>> _get_python_type_with_dimensions([2], "float")
    'tuple[float, float]'
    >>> _get_python_type_with_dimensions(["n"], "int")
    'tuple[int, ...]'
    >>> _get_python_type_with_dimensions([2, 3], "float")
    'tuple[tuple[float, float, float], tuple[float, float, float]]'
    >>> _get_python_type_with_dimensions([2, 1], "float")
    'tuple[tuple[float], tuple[float]]'
    >>> _get_python_type_with_dimensions(["n", "m", 2], "Quantity")
    'tuple[tuple[tuple[Quantity, Quantity], ...], ...]'
    """

    def _dimension(inner_type: str, dimension_len: Union[int, str]) -> str:
        if isinstance(dimension_len, str):
            content = [inner_type, "..."]
        else:
            content = [inner_type] * dimension_len

        return f"tuple[{', '.join(content)}]"

    # Start with base
    res = base

    # then wrap dimension information on top as it applies
    for dim in dimension_info[::-1]:
        res = _dimension(res, dim)

    return res


@define
class FortranDataType:
    """
    A fortran data type including its attributes

    See the docstring of :mod:`fgen.fortran_parsing`
    for further details on Fortran type declaration
    statements and how we handle and describe them.

    We are only interested in the type attribute declaration
    so anything including and after a "::" should be stripped
    before initialising this class.
    """

    type_specification: str
    """
    This can be either an intrinsic type, or a derived type.

    For example, "real(8)", "logical", "type(my_calculator)"
    """

    attribute_specifications: tuple[
        Union[str, DimensionAttributeSpecification], ...
    ] = field()
    """
    List of attribute specifications that apply to the Fortran data type

    These are our internal representation of attribute specifications
    e.g. "pointer", "target", "dimension(3, 3)"
    """

    def __str__(self) -> str:
        return self.fortran_type_attribute_declaration

    @attribute_specifications.validator
    def _check_attribute_specifications(
        self,
        attribute: Any,
        value: tuple[Union[str, DimensionAttributeSpecification], ...],
    ) -> None:
        dimension_count = sum(
            isinstance(a, DimensionAttributeSpecification) for a in value
        )
        if dimension_count > 1:
            raise ValueError(  # noqa: TRY003
                "More than one dimension attribute specification "
                "(can Fortran be compiled "
                "with more than one dimension attribute specification?)"
            )

    @classmethod
    def from_str(cls, fortran_type_attribute_declaration: str) -> FortranDataType:
        """
        Create :obj:`FortranDataType` from type declaration statement

        Parses and validates a type declaration statement

        Parameters
        ----------
        fortran_type_attribute_declaration
            Fortran type declaration statement

            For example, "real(8), dimension(2) :: heat_uptake_sensitivity"

            The entity declaration component (including and after "::")
            is ignored hence is optional.

            See the docstring of :mod:`fgen.fortran_parsing` for further details.

        Raises
        ------
        ValueError
            The type declaration statement is invalid
        """
        type_specification, attribute_specifications = _get_parts(
            fortran_type_attribute_declaration
        )
        _validate_fortran_type_attribute_declaration(
            type_specification, attribute_specifications
        )

        return cls(
            type_specification=type_specification,
            attribute_specifications=_convert_complex_attribute_specifications(
                attribute_specifications
            ),
        )

    @property
    def fortran_type_attribute_declaration(self) -> str:
        """
        Fortran type attribute declaration, including attribute specifications.

        Attribute specifications are only included if they are applicable.

        See the docstring of :mod:`fgen.fortran_parsing` for further details.
        """
        if not self.attribute_specifications:
            return self.type_specification

        attribute_specifications = ", ".join(
            str(item) for item in self.attribute_specifications
        )
        return f"{self.type_specification}, {attribute_specifications}"

    @property
    def python_equivalent_type_annotation(self) -> str:
        """
        Type annotation for the python-equivalent of this Fortran type
        """
        base_type = self._base_python_type()
        dimension_attribute_specification = self.dimension_attribute_specification
        if not dimension_attribute_specification:
            return base_type

        # TODO: Remove this.
        # This custom handling of shapes will be silly if we have an array of say,
        # size 100, so it makes more sense to just use numpy return types here.
        return _get_python_type_with_dimensions(
            dimension_attribute_specification.dimensions, base_type
        )

    @property
    def is_array(self) -> bool:
        """
        Whether this instance represents an array or not.

        Returns
        -------
            ``True`` if self represents an array, ``False`` otherwise.
        """
        return self.dimension_attribute_specification is not None

    @property
    def has_deferred_size(self) -> bool:
        """
        Whether this instance represents a type with a deferred size or not.

        Returns
        -------
            ``True`` if self represents a type with a deferred size,
            ``False`` otherwise.
        """
        return ":" in self.fortran_type_attribute_declaration

    @property
    def is_character(self) -> bool:
        """
        Whether this instance represents a character type

        Returns
        -------
            ``True`` if self represents a character type, ``False`` otherwise.
        """
        return self.type_specification.startswith("character")

    @property
    def is_logical(self) -> bool:
        """
        Whether this instance represents a logical type

        Returns
        -------
            ``True`` if self represents a logical type, ``False`` otherwise.
        """
        return self.type_specification.startswith("logical")

    @property
    def is_float_double(self) -> bool:
        """
        Whether this instance represents a type which is a floating double-point type.

        Returns
        -------
            ``True`` if self represents a type which is a floating double-point type,
            ``False`` otherwise.
        """
        return "real" in self.type_specification and "8" in self.type_specification

    @property
    def is_derived_type(self) -> bool:
        """
        Whether this instance represents a type which is a derived type

        Returns
        -------
            ``True`` if self represents a type which is a derived type
            ``False`` otherwise.
            True if a type is a derived type, False if it is an intrinsic type
        """
        return self.type_specification.startswith("type(")

    @property
    def is_pointer(self) -> bool:
        """
        Whether this instance represents a type which is a pointer

        Returns
        -------
            ``True`` if self represents a type which is a pointer
            ``False`` otherwise.
        """
        return "pointer" in self.attribute_specifications

    @property
    def is_deferred_array(self) -> bool:
        """
        Whether this instance represents a deferred size array

        Returns
        -------
            ``True`` if self represents a type which is a deferred size array
            ``False`` otherwise.
        """
        dim_attrs = self.dimension_attribute_specification

        if dim_attrs is None:
            return False
        return ":" in dim_attrs.dimensions

    @property
    def is_enum(self) -> bool:
        """
        Whether this instance represents a type which is an enum.

        Returns
        -------
            ``True`` if self represents a type which is an enum.
            ``False`` otherwise.
        """
        # Hmm, not sure whether integer(kind(8)) is valid Fortran
        # and would trip this up....
        return (
            "integer" in self.type_specification and "kind(" in self.type_specification
        )

    @property
    def equivalent_python_type(self) -> str:
        """
        Python type-hint for the data type

        The "base" type of the equivalent Python type will be:

        If the Fortran type has a "dimension" attribute specification,
        this modifies the "base" type to a `tuple`
        that respects the intended shape of the value.
        n-dimensional variables are supported.
        TODO: remove this behaviour, doesn't scale well to long arrays
        and doesn't help with deferred shape arrays.
        """
        base_type = self._base_python_type()
        dimension_attribute_specification = self.dimension_attribute_specification

        if base_type != "Quantity" and dimension_attribute_specification:
            # Pint doesn't support dimension attributes
            # so we only do this if the Python type is not a pint type
            return _get_python_type_with_dimensions(
                dimension_attribute_specification.dimensions, base_type
            )

        return base_type

    def _base_python_type(self) -> str:
        """
        Determine the "base" Python type that represents the Fortran type declaration

        This focuses on the base type (float, int etc.)
        and ignores the modification of this type from attributes such as "dimension"
        (which mean the 'full' python type is some sort of array or iterable).
        Such dimension handling is done in :attr:`~equivalent_python_type`.

        Returns
        -------
            The string representation of the python type
            that is equivalent to the base type
            (i.e. ignoring any array or other container) of ``self``
        """
        if self.is_derived_type:
            match = re.match(
                _DERIVED_TYPE_REGEX, self.type_specification, flags=re.IGNORECASE
            )
            if not match:
                raise ValueError(  # noqa: TRY003 # pragma: no cover # should be impossible to get here
                    f"Could not determine python type for {self.type_specification}"
                )

            return match.group("derived_type_name").strip()

        else:
            tsl = self.type_specification.lower()

            if tsl.startswith("real"):
                return "float"

            if tsl.startswith("integer"):
                match = re.match(
                    _ENUM_LIKE_REGEX, self.type_specification, flags=re.IGNORECASE
                )
                if match:
                    return match.group("enum_name").strip()

                return "int"

            if tsl.startswith("character"):
                return "str"

            if tsl.startswith("logical"):
                return "bool"

            if tsl.startswith("complex"):
                msg = (
                    "We have not yet worked out how to support "
                    f"{self.type_specification}, please raise an issue"
                )
                raise NotImplementedError(msg)

            raise ValueError(  # pragma: no cover # not sure how you could reach here
                self.type_specification
            )

    @property
    def dimension_attribute_specification(
        self,
    ) -> Optional[DimensionAttributeSpecification]:
        """
        Get the dimension attribute of the type if it exists

        Returns
        -------
            The dimension attribute if it is present otherwise None
        """
        for attr in self.attribute_specifications:
            if isinstance(attr, DimensionAttributeSpecification):
                # there can only be one dimension attribute so as soon as we
                # find it, we can stop looking
                return attr

        return None



"""
Generation of the Python init module
"""
from __future__ import annotations

from pathlib import Path

from fgen.data_models import (
    Package,
)
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
    post_process_jinja_rendering,
)
from fgen.wrapper_building.formatting import format_python_code


def generate_python_init_module(package: Package) -> str:
    """
    Generate the Python init module

    Parameters
    ----------
    package
        Package in which the module lives

    Returns
    -------
        Python init file as code
    """
    # Not using the builder pattern for this module due to how simple the output is

    template = get_template_in_directory(
        "python-init-module.py.jinja", Path(__file__).parent, JINJA_ENV
    )

    result = post_process_jinja_rendering(template.render())

    return format_python_code(result)



"""
Data models

These data models describe the configuration that is used by this package
and the schema used to serialize and validate any configuration that is loaded.
For a more detailed description, see :ref:`overview-reference`.

Each module that is being wrapped is represented
using a :class:`~fgen.data_models.module.Module`
and can be loaded from disk using
:func:`~fgen.data_models.serialisation.load_module_definition`.
"""

from __future__ import annotations

from fgen.data_models.fortran_derived_type import FortranDerivedType
from fgen.data_models.method import Method
from fgen.data_models.module import Module
from fgen.data_models.module_enum_defining import ModuleEnumDefining
from fgen.data_models.module_requirement import ModuleRequirement
from fgen.data_models.multi_return import MultiReturn
from fgen.data_models.package import Package
from fgen.data_models.package_shared_elements import PackageSharedElements
from fgen.data_models.serialisation import (
    converter,
    load_enum_defining_module,
    load_module_definition,
)
from fgen.data_models.unitless_value import UnitlessValue
from fgen.data_models.value import Value

__all__ = [
    "FortranDerivedType",
    "Method",
    "Module",
    "ModuleEnumDefining",
    "ModuleRequirement",
    "MultiReturn",
    "Package",
    "PackageSharedElements",
    "Value",
    "UnitlessValue",
    "converter",
    "load_enum_defining_module",
    "load_module_definition",
]



"""
Data model of a method
"""
from __future__ import annotations

from typing import Union

from attrs import define

from fgen.data_models.multi_return import MultiReturn
from fgen.data_models.value import Value


@define
class Method:
    """
    Data model of a Fortran method

    This method is assumed to be part of a
    :class:`~fgen.data_models.fortran_derived_type.FortranDerivedType`.

    This function can have multiple :attr:`parameters`
    and a single, optional :attr:`returns` value.
    """

    name: str
    """Name of the function in Fortran"""

    description: str
    """Description of what the function does"""

    parameters: dict[str, Value]
    """
    Collection of named parameters of the function

    Currently, only :obj:`ValueDefinition` is supported because
    we assume that each parameter only refers to
    one value. If you think this assumption should be
    relaxed, please raise an issue illustrating your use
    case.
    """

    returns: Union[Value, MultiReturn]
    """Return value of the function"""

    @property
    def units(self) -> dict[str, Union[str, tuple[str, ...]]]:
        """
        Units used in the method

        Includes units from of the parameters and the return value.

        Raises
        ------
        ValueError
            If different units are supplied for a given parameter name

        Returns
        -------
            Collection of parameter names and associated units
        """
        res: dict[str, Union[str, tuple[str, ...], None]] = {
            key: parameter.unit for key, parameter in self.parameters.items()
        }

        if self.returns:
            rn = self.returns.definition.name  # rn = return's name
            if rn in res and res[rn] != self.returns.unit:
                raise ValueError(  # noqa: TRY003
                    f"Inconsistent units for attribute '{rn}'. "
                    f"In the input parameters it has units '{res[rn]}' "
                    f"whereas in the returns it has units '{self.returns.unit}'"
                )

            res[rn] = self.returns.unit

        # Drop any None units
        return {k: v for k, v in res.items() if v is not None}



"""
Information injection functions
"""
from __future__ import annotations

from typing import Union

from fgen.data_models import MultiReturn, Value
from fgen.wrapping_strategies.interface import WrappingStrategyLike


def inject_wrapping_strategy_information(
    rendered_code: str,
    value: Union[MultiReturn, Value],
    wrapping_strategy: WrappingStrategyLike,
    comment_character: str,
    indent: str = "",
) -> str:
    """
    Inject information about the strategy as a comment

    Parameters
    ----------
    rendered_code
        Rendered code, into which to inject the information

    value
        Value being wrapped

    wrapping_strategy
        Wrapping strategy used

    comment_character
        Comment character for this text

    indent
        Indent to use when starting a new line

    Returns
    -------
        Rendered code, with wrapping strategy information injected
    """
    ws_repr_raw = repr(wrapping_strategy)
    toks = ws_repr_raw.split("(")

    wrapping_strategy_name = toks[0]
    wrapping_strategy_attribute_values = toks[1][:-1].split(",")
    comment = f"\n{indent}".join(
        [
            f"{comment_character} Wrapping {value.definition.name}",
            f"{comment_character} Strategy: {wrapping_strategy_name}(",
            *[
                f"{comment_character}     {v.strip()},"
                for v in wrapping_strategy_attribute_values
                if v.strip()
            ],
            f"{comment_character} )",
        ]
    )

    return "\n".join([comment, rendered_code])



"""
Data model for how we pass data from Python to Fortran
"""
from __future__ import annotations

from typing import Optional

from attrs import define


@define
class PassingToFortranSteps:
    """
    Container for holding the steps required to pass data to Fortran
    """

    preparatory_python_calls: Optional[str]
    """Calls required in Python to prepare to pass the data to Fortran"""

    fortran_module_callable_args: tuple[tuple[str, str], ...]
    """
    Tuple of argument-value pairs to pass to the Fortran module callable

    The first element of each element in ``fortran_module_callable_args``
    is the name expected by Fortran,
    the second element is the name of the Python variable.

    There might be differences from the user-facing Python arguments,
    because we don't always pass values straight through.
    """

    @property
    def fortran_callable_arg_list(self) -> list[str]:
        """
        Argument list to use when calling a Fortran callable

        Note that this may not be the entire argument list,
        it is only the calls for the arguments managed by ``self``.
        """
        return [
            f"{arg_fortran}={arg_python}"
            for arg_fortran, arg_python in self.fortran_module_callable_args
        ]



"""
Generation of the Python enums module

This holds the Python equivalent of any enums defined in Fortran.
"""
from __future__ import annotations

from pathlib import Path

from fgen.data_models import Package
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
    post_process_jinja_rendering,
)
from fgen.wrapper_building.formatting import format_python_code


def generate_python_enums_module(
    package: Package,
) -> str:
    """
    Generate the Python enums module

    Parameters
    ----------
    package
        Package for which to generate the Python enums module

    Returns
    -------
        Python enums module
    """
    # Not using the builder pattern for this module due to how simple the output is

    template = get_template_in_directory(
        "python-enums-module.py.jinja", Path(__file__).parent, JINJA_ENV
    )
    template_enum = get_template_in_directory(
        "python-enum-definition.py.jinja", Path(__file__).parent, JINJA_ENV
    )

    enums = []
    for module_enum_defining in package.modules_enum_defining:
        enums.append(
            template_enum.render(enum_definition=module_enum_defining.provides)
        )

    result = post_process_jinja_rendering(template.render(enums=enums))

    return format_python_code(result)



"""
Wrapping strategies

Different Fortran types need different strategies for wrapping.
We capture these here.
"""
from __future__ import annotations

from fgen.fortran_parsing import FortranDataType
from fgen.wrapping_strategies.array_deferred_size import (
    WrappingStrategyArrayDeferredSize,
)
from fgen.wrapping_strategies.character import (
    WrappingStrategyCharacter,
)
from fgen.wrapping_strategies.character_deferred_size import (
    WrappingStrategyCharacterDeferredSize,
)
from fgen.wrapping_strategies.default import WrappingStrategyDefault
from fgen.wrapping_strategies.derived_type import (
    WrappingStrategyDerivedType,
)
from fgen.wrapping_strategies.enum import WrappingStrategyEnum
from fgen.wrapping_strategies.interface import WrappingStrategyLike
from fgen.wrapping_strategies.logical import (
    WrappingStrategyLogical,
)
from fgen.wrapping_strategies.passing_to_fortran_steps import (
    PassingToFortranSteps,
)


def get_wrapping_strategy(  # noqa: PLR0911
    fortran_data_type: FortranDataType,
) -> WrappingStrategyLike:
    """
    Get wrapping strategy for a given :obj:`FortranDataType`

    Parameters
    ----------
    fortran_data_type
        :obj:`FortranDataType` for which to get the wrapping strategy

    Returns
    -------
        Wrapping strategy object
    """
    if fortran_data_type.is_derived_type:
        return WrappingStrategyDerivedType()

    if fortran_data_type.is_array and fortran_data_type.has_deferred_size:
        return WrappingStrategyArrayDeferredSize()

    if fortran_data_type.is_character:
        if fortran_data_type.has_deferred_size:
            return WrappingStrategyCharacterDeferredSize()
        else:
            return WrappingStrategyCharacter()

    if fortran_data_type.is_logical:
        return WrappingStrategyLogical()

    if fortran_data_type.is_enum:
        return WrappingStrategyEnum()

    return WrappingStrategyDefault()


__all__ = [
    "PassingToFortranSteps",
    "WrappingStrategyArrayDeferredSize",
    "WrappingStrategyCharacter",
    "WrappingStrategyCharacterDeferredSize",
    "WrappingStrategyDefault",
    "WrappingStrategyDerivedType",
    "WrappingStrategyLike",
    "WrappingStrategyLogical",
    "get_wrapping_strategy",
]



"""
Data model for how we receive data from Python in Fortran
"""
from __future__ import annotations

from typing import Optional, Union

from attrs import define

from fgen.data_models import MultiReturn, Value


@define
class ReceivingFromPythonSteps:
    """
    Container for holding the steps required to receive data from Python
    """

    postprocessing_fortran_calls: Optional[str]
    """Calls required to process the data received from Python"""

    fortran_wrapper_callable_parameters: tuple[
        tuple[str, str, Union[MultiReturn, Value]], ...
    ]
    """
    Parameters of the Fortran wrapper callable

    The first element of each tuple is the parameter name
    as it appears in the declaration of the Fortran wrapper callable.
    The second element of each tuple is the type used
    when passing this variable to Fortran
    (which may differ from the variable type in Fortran).
    The third element of each tuple is the data model of the parameter,
    which makes it available for docstrings etc.
    """

    fortran_module_callable_args: tuple[Union[MultiReturn, Value], ...]
    """
    Arguments to pass to the (original) Fortran module callable

    These just take the data model of the (original) Fortran module callable's argument.
    Hence these must match what is expected by the Fortran module exactly,
    there can be no different naming, types etc.
    (different types wouldn't compile anyway,
    avoiding different names is done for clarity and simplicity).
    """

    @property
    def fortran_wrapper_callable_parameter_names(self) -> list[str]:
        """
        Names of the Fortran wrapper callable's parameters

        Note that this may not be the entire parameter list,
        it is only the calls for the parameters managed by ``self``.
        """
        return [v[0] for v in self.fortran_wrapper_callable_parameters]

    @property
    def fortran_module_callable_args_names(self) -> list[str]:
        """
        Argument list to use when calling a callable in the (original) Fortran module

        Note that this may not be the entire argument list,
        it is only the calls for the arguments managed by ``self``.
        """
        return [v.definition.name for v in self.fortran_module_callable_args]

    @property
    def not_directly_passed_fortran_module_callable_args_type_declarations(
        self,
    ) -> list[tuple[str, str]]:
        """
        Type declarations for the argument list that aren't directly passed from Python.

        These type declarations are only required
        because these arguments for the (original) Fortran callable
        cannot be directly passed from Python,
        i.e. these arguments are not parameters of the wrapper module's callable
        (i.e. would otherwise be undefined).

        The first element of each returned tuple is the argument's name,
        the second is the type declaration.

        The argument list is the argument list that is used
        when calling a callable in the (original) Fortran module.

        Note that this may not be the entire argument list,
        it is only the calls for the arguments managed by ``self``.
        """
        out = [
            (
                v.definition.name,
                get_type_declaration_for_passing_value_to_fortran_callable(v),
            )
            for v in self.not_directly_passed_fortran_module_callable_args
        ]

        return out

    @property
    def not_directly_passed_fortran_module_callable_args(
        self,
    ) -> list[Union[MultiReturn, Value]]:
        """
        Fortran module callable arguments that aren't directly passed from Python

        These arguments will have intermediate steps,
        hence the arguments need to be declared as variables separately
        from the callable parameter declarations.

        Note that this is only for the arguments managed by ``self``,
        not necessarily all arguments of the callable.
        """
        out = [
            v
            for v in self.fortran_module_callable_args
            if v.definition.name not in self.fortran_wrapper_callable_parameter_names
        ]

        return out


def get_type_declaration_for_passing_value_to_fortran_callable(
    value: Union[MultiReturn, Value],
) -> str:
    """
    Get the type declaration for ``value`` when passing to the (original) Fortran.

    This is the type expected by the (original) Fortran
    when receiving ``value`` as an argument to callables.

    Parameters
    ----------
    value
        Value for which to get the type declaration

    Returns
    -------
        The type declaration for ``value``
    """
    # Lazy import to avoid circular dependency
    from fgen.wrapping_strategies import get_wrapping_strategy

    wrapping_strategy = get_wrapping_strategy(value.definition.as_fortran_data_type())

    return wrapping_strategy.get_type_attribute_declaration_for_value_to_pass_to_original_fortran_module(  # noqa: E501
        value
    )



"""
Wrapping strategy for everything that isn't covered elsewhere, i.e. the default

This works for basic intrinsic types (floats and integers)
and can be used when we haven't implemented something more specific
(probably with iterations required thereafter
because the default strategy is unlikely to just work out of the box).
"""
from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Optional, Union

from attrs import define

from fgen.data_models import Method, Module, MultiReturn, Value
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
)
from fgen.wrapping_strategies.information_injection import (
    inject_wrapping_strategy_information,
)
from fgen.wrapping_strategies.passing_to_fortran_steps import (
    PassingToFortranSteps,
)
from fgen.wrapping_strategies.receiving_from_python_steps import (
    ReceivingFromPythonSteps,
)
from fgen.wrapping_strategies.type_annotations import (
    PINT_QUANTITY_TYPE_ANNOTATION,
)


@define
class WrappingStrategyDefault:
    """
    Wrapping strategy for everything that isn't covered elsewhere, i.e. the default
    """

    magnitude_suffix: str = "_m"
    """Suffix to add to the variable name to indicate that this is the magnitude"""

    def get_fortran_wrapper_module_getter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iget_{value.definition.name}"

    def get_fortran_wrapper_module_setter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iset_{value.definition.name}"

    def get_fortran_wrapper_statement_declarations_getters_and_setters(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> list[str]:
        return [
            f"public :: {method_name}"
            for method_name in (
                self.get_fortran_wrapper_module_getter(value),
                self.get_fortran_wrapper_module_setter(value),
            )
        ]

    def get_fortran_wrapper_statement_declarations_for_returning_method_result_to_python(  # noqa: E501, D102
        self, method: Method, prefix: str
    ) -> list[str]:
        return [f"public :: {prefix}{method.name}"]

    def get_receiving_from_python_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str] = None,
        providing_module: Optional[Module] = None,
        get_instance_callable_name: Optional[str] = None,
    ) -> ReceivingFromPythonSteps:
        if value.dynamic_unit:
            # The two paths are currently the same.
            # It would be better if the fortran module callable
            # received an argument with name
            # f"{value.definition.name}{self.magnitude_suffix}",
            # to make clear that we only pass in the magniutude.
            # This would require changes in a few places.
            return ReceivingFromPythonSteps(
                postprocessing_fortran_calls=None,
                fortran_wrapper_callable_parameters=(
                    (
                        value.definition.name,
                        value.definition.as_fortran_data_type().fortran_type_attribute_declaration,
                        value,
                    ),
                ),
                fortran_module_callable_args=(value,),
            )

        return ReceivingFromPythonSteps(
            postprocessing_fortran_calls=None,
            fortran_wrapper_callable_parameters=(
                (
                    value.definition.name,
                    value.definition.as_fortran_data_type().fortran_type_attribute_declaration,
                    value,
                ),
            ),
            fortran_module_callable_args=(value,),
        )

    def get_type_attribute_declaration_for_value_to_pass_to_original_fortran_module(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return (
            value.definition.as_fortran_data_type().fortran_type_attribute_declaration
        )

    def get_fortran_for_getter(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        value_manager_get_free_instance_number: Optional[str] = None,
        value_manager_get_instance: Optional[str] = None,
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-getter.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            return_value=value,
            fortran_wrapper_module_callable=self.get_fortran_wrapper_module_getter(
                value
            ),
            class_being_wrapped=class_being_wrapped,
            fortran_module_attribute=value.definition.name,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="!"
        )

    def get_fortran_for_method_returning_wrapped_type(  # noqa: D102, PLR0913
        self,
        fortran_wrapper_module_callable: str,
        receiving_from_python_steps: ReceivingFromPythonSteps,
        return_value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        method_name: str,
        return_value_manager_get_free_instance_number: Optional[str],
        return_value_manager_get_instance: Optional[str],
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-method-returning-wrapped-type.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        input_params = (
            receiving_from_python_steps.fortran_wrapper_callable_parameter_names
        )

        result = template.render(
            fortran_wrapper_module_callable=fortran_wrapper_module_callable,
            return_value=return_value,
            input_params=input_params,
            receiving_from_python_steps=receiving_from_python_steps,
            class_being_wrapped=class_being_wrapped,
            method_name=method_name,
        )

        return inject_wrapping_strategy_information(
            result, value=return_value, wrapping_strategy=self, comment_character="!"
        )

    def get_python_user_facing_name(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.name

    def get_python_post_verify_units_input_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        if value.dynamic_unit:
            return PINT_QUANTITY_TYPE_ANNOTATION

        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_return_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        if value.dynamic_unit:
            return PINT_QUANTITY_TYPE_ANNOTATION

        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_argument_declaration_type_annotation(  # noqa: D102
        self, value: Value
    ) -> str:
        if value.dynamic_unit:
            return PINT_QUANTITY_TYPE_ANNOTATION

        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_getter_docstring(self, value: Value) -> str:  # noqa: D102
        template = get_template_in_directory(
            "python-getter-docstring.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(attribute=value)

        return result

    def get_python_setter_docstring(self, value: Value) -> str:  # noqa: D102
        return f"Setter for ``{value.definition.name}``"

    def generate_python_for_fortran_return_value_processing(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        fortran_module_callable: str,
        fortran_module_callable_args: Iterable[str],
        dynamic_unit: Optional[str],
    ) -> str:
        return_type = self.get_python_return_type_annotation(value)

        template = get_template_in_directory(
            "python-for-fortran-return-value-processing.py.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            result_name=self.get_python_user_facing_name(value),
            fortran_module_callable=fortran_module_callable,
            fortran_module_callable_args=fortran_module_callable_args,
            return_type=return_type,
            dynamic_unit=dynamic_unit,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="#"
        )

    def get_passing_to_fortran_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str],
    ) -> PassingToFortranSteps:
        if value.dynamic_unit:
            argument_magnitude = f"{value.definition.name}{self.magnitude_suffix}"

            template = get_template_in_directory(
                "python-for-preparing-to-pass-to-fortran-dynamic-unit.py.jinja",
                Path(__file__).parent,
                JINJA_ENV,
            )
            preparatory_python_calls = template.render(
                argument_user_facing=self.get_python_user_facing_name(value),
                argument_magnitude=argument_magnitude,
                dynamic_unit_source=dynamic_unit,
            )

            return PassingToFortranSteps(
                preparatory_python_calls=inject_wrapping_strategy_information(
                    preparatory_python_calls,
                    value=value,
                    wrapping_strategy=self,
                    comment_character="#",
                ),
                # TODO: re-name this to fortran_wrapper_callable_args
                # because we are calling the wrapper here, not the original module.
                fortran_module_callable_args=(
                    (value.definition.name, argument_magnitude),
                ),
            )

        return PassingToFortranSteps(
            preparatory_python_calls=None,
            fortran_module_callable_args=(
                (value.definition.name, self.get_python_user_facing_name(value)),
            ),
        )



"""
Base class for wrapped derived types
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from functools import wraps
from types import TracebackType
from typing import Any, Callable, Optional, TypeVar

import attrs
from attrs import define, field
from typing_extensions import Concatenate, ParamSpec

from fgen_runtime.exceptions import InitialisationError
from fgen_runtime.units import FuncT

INVALID_INSTANCE_INDEX: int = -1
"""
Value used to denote an invalid ``instance_index``.

This can occur value when a wrapper class
has not yet been initialised (connected to a Fortran instance).
"""


@define
class FinalizableWrapperBase(ABC):
    """
    Base class for model wrappers
    """

    instance_index: int = field(
        validator=attrs.validators.instance_of(int),
        default=INVALID_INSTANCE_INDEX,
    )
    """
    Model index of wrapper Fortran instance
    """

    @classmethod
    @abstractmethod
    def from_new_connection(cls) -> FinalizableWrapperBase:
        """
        Initialise by establishing a new connection with the Fortran module

        This requests a new model index from the Fortran module and then
        initialises a class instance

        Returns
        -------
        New class instance
        """
        ...

    @abstractmethod
    def finalize(self) -> None:
        """
        Finalise the Fortran instance and set self back to being uninitialised

        This method resets ``self.instance_index`` back to
        ``_UNINITIALISED_instance_index``

        Should be decorated with :func:`check_initialised`
        """
        # call to Fortran module goes here when implementing
        self._uninitialise_instance_index()

    def _uninitialise_instance_index(self) -> None:
        self.instance_index = INVALID_INSTANCE_INDEX

    @property
    def initialized(self) -> bool:
        """
        Is the instance initialised, i.e. connected to a Fortran instance?
        """
        return self.instance_index != INVALID_INSTANCE_INDEX


@define
class FinalizableWrapperBaseContext(AbstractContextManager):  # type: ignore
    """
    Context manager for a wrapper
    """

    model: FinalizableWrapperBase
    """
    model instance to be managed
    """

    def __enter__(self) -> FinalizableWrapperBase:
        if not self.model.initialized:
            raise InitialisationError(self.model)

        return self.model

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.model.finalize()


P = ParamSpec("P")
T = TypeVar("T")
Wrapper = TypeVar("Wrapper", bound=FinalizableWrapperBase)


def check_initialised(method: FuncT) -> FuncT:
    """
    Check that initialisation is called before execution

    Parameters
    ----------
    method
        Method to wrap

    Returns
    -------
    Wrapped method. Before the method is called, initialisation will first be
    checked. If the model wrapper hasn't been initialised before the method is
    called, an ``InitialisationError`` will be raised.
    """

    @wraps(method)
    def checked(
        ref: Wrapper,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if not ref.initialized:
            raise InitialisationError(ref, method)

        return method(ref, *args, **kwargs)

    return checked  # type: ignore


# Thank you for type hints info
# https://adamj.eu/tech/2021/05/11/python-type-hints-args-and-kwargs/
def execute_finalize_on_fail(
    inst: FinalizableWrapperBase,
    func_to_try: Callable[Concatenate[int, P], T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    """
    Execute a function, finalising the instance before raising if any error occurs

    This function is most useful in factory functions where it provides a
    clean way of ensuring that any Fortran is freed up in the event of an
    initialisation failure for any reason

    Parameters
    ----------
    inst
        Instance whose model index we will use when executing the functin

    func_to_try
        Function to try executing, must take ``inst``'s model index as its
        first argument

    *args
        Passed to ``func_to_try``

    **kwargs
        Passed to ``func_to_try``

    Returns
    -------
    Result of calling ``func_to_try(inst.instance_index, *args, **kwargs)``

    Raises
    ------
    Exception
        Any exception which occurs when calling ``func_to_try. Before the
        exception is raised, ``inst.finalize()`` is called.
    """
    try:
        return func_to_try(inst.instance_index, *args, **kwargs)
    except RuntimeError:
        # finalize the instance before raising
        inst.finalize()

        raise



"""
Generation of the Fortran manager module
"""

from pathlib import Path

from attrs import define

from fgen.data_models import Module, Package, PackageSharedElements
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
    post_process_jinja_rendering,
)


@define
class FortranManagerModuleBuilder:
    """
    Builder of Fortran manager modules
    """

    package: Package
    """
    Package for which the builder is building wrappers
    """

    module: Module
    """
    Module for which to build the wrapper
    """

    shared: PackageSharedElements
    """
    Elements which have to be shared across the package

    For example, the names of functions which are used in more than one wrapper module.
    """


def generate_fortran_manager_module(builder: FortranManagerModuleBuilder) -> str:
    """
    Generate the Fortran manager module

    Parameters
    ----------
    builder
        Builder to use to generate the Fortran manager module

    Returns
    -------
        Fortran manager module as code
    """
    template = get_template_in_directory(
        "fortran-manager-module.f90.jinja", Path(__file__).parent, JINJA_ENV
    )

    result = post_process_jinja_rendering(template.render(builder=builder))

    return result



"""
Unit-handling
"""
from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Callable, TypeVar, Union, overload

import numpy as np
import numpy.typing as npt
import pint
from typing_extensions import ParamSpec, TypeAlias

UnitDefinition: TypeAlias = Union[str, pint.registry.Unit]
"""Types which can be used to specify the exepcted (input) unit"""

ReturnUnitDefinition: TypeAlias = Union[
    str, tuple[str, ...], pint.registry.Unit, tuple[pint.registry.Unit, ...]
]
"""Types which can be used to specify the exepcted return unit"""

VerifyUnitsSupported: TypeAlias = Union[UnitDefinition, None]
"""Types which are supported as inputs to :func:`verify_units`"""

Q: TypeAlias = pint.registry.UnitRegistry.Quantity
"""Short-hand to save space and typing"""

PintSupported: TypeAlias = Union[
    float, int, np.number[Any], npt.NDArray[np.number[Any]], npt.ArrayLike
]
"""
Types that are able to be cast to :obj:`pint.registry.UnitRegistry.Quantity` by pint
"""

P = ParamSpec("P")
FuncT = TypeVar("FuncT", bound=Callable[..., Any])


# Overload variants created with scripts/make-verify-units-overloads.py
# Best not to do this by hand unless you want cramp

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
X = TypeVar("X")
Y = TypeVar("Y")
Z = TypeVar("Z")
UT = TypeVar("UT")
VT = TypeVar("VT")
PintT = TypeVar("PintT", bound=PintSupported)
PintU = TypeVar("PintU", bound=PintSupported)
PintV = TypeVar("PintV", bound=PintSupported)
PintX = TypeVar("PintX", bound=PintSupported)
PintY = TypeVar("PintY", bound=PintSupported)
PintZ = TypeVar("PintZ", bound=PintSupported)
PintUT = TypeVar("PintUT", bound=PintSupported)
PintVT = TypeVar("PintVT", bound=PintSupported)


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[()],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[P, PintT]], Callable[P, Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU], PintT]], Callable[[Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[P, PintT]], Callable[P, Q]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU], T]], Callable[[Q], T]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV], PintT]], Callable[[Q, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U], PintT]], Callable[[Q, U], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU], PintT]], Callable[[U, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[P, PintT]], Callable[P, Q]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV], T]], Callable[[Q, Q], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U], T]], Callable[[Q, U], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU], T]], Callable[[U, Q], T]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV, PintX], PintT]], Callable[[Q, Q, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV, U], PintT]], Callable[[Q, Q, U], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U, PintV], PintT]], Callable[[Q, U, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, None, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U, V], PintT]], Callable[[Q, U, V], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV], PintT]], Callable[[U, Q, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, V], PintT]], Callable[[U, Q, V], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, V, PintU], PintT]], Callable[[U, V, Q], Q]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, None, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[P, PintT]], Callable[P, Q]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV, PintX], T]], Callable[[Q, Q, Q], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV, U], T]], Callable[[Q, Q, U], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U, PintV], T]], Callable[[Q, U, Q], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, None, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, U, V], T]], Callable[[Q, U, V], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV], T]], Callable[[U, Q, Q], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, UnitDefinition, None],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, V], T]], Callable[[U, Q, V], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, None, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, V, PintU], T]], Callable[[U, V, Q], T]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY], PintT]], Callable[[Q, Q, Q, Q], Q]
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV, PintX], PintT]], Callable[[U, Q, Q, Q], Q]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[PintU, PintV, PintX, PintY], T]], Callable[[Q, Q, Q, Q], T]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV, PintX], T]], Callable[[U, Q, Q, Q], T]]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ], PintT]], Callable[[Q, Q, Q, Q, Q], Q]
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[None, UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY], PintT]], Callable[[U, Q, Q, Q, Q], Q]
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ], T]], Callable[[Q, Q, Q, Q, Q], T]
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[None, UnitDefinition, UnitDefinition, UnitDefinition, UnitDefinition],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY], T]], Callable[[U, Q, Q, Q, Q], T]
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ, PintUT], PintT]],
    Callable[[Q, Q, Q, Q, Q, Q], Q],
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY, PintZ], PintT]],
    Callable[[U, Q, Q, Q, Q, Q], Q],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ, PintUT], T]],
    Callable[[Q, Q, Q, Q, Q, Q], T],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY, PintZ], T]],
    Callable[[U, Q, Q, Q, Q, Q], T],
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ, PintUT, PintVT], PintT]],
    Callable[[Q, Q, Q, Q, Q, Q, Q], Q],
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY, PintZ, PintUT], PintT]],
    Callable[[U, Q, Q, Q, Q, Q, Q], Q],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[PintU, PintV, PintX, PintY, PintZ, PintUT, PintVT], T]],
    Callable[[Q, Q, Q, Q, Q, Q, Q], T],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, PintY, PintZ, PintUT], T]],
    Callable[[U, Q, Q, Q, Q, Q, Q], T],
]:
    ...


# Below here are overloads written by hand
# Our lack of all combos means we have to write some by hand.
# This could become extremely painful, or this file may become ludicrously long.
@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        None,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV, V], X]], Callable[[U, Q, Q, V], X]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        None,
        None,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[[U, PintU, PintV, V, X], Y]], Callable[[U, Q, Q, V, X], Y]]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        None,
        UnitDefinition,
        UnitDefinition,
        None,
        None,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, V, PintY, PintZ, X, Y], Z]],
    Callable[[U, Q, Q, Q, V, Q, Q, X, Y], Z],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        None,
        UnitDefinition,
        UnitDefinition,
        None,
        None,
        None,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, PintU, PintV, PintX, V, PintY, PintZ, X, Y, Z], UT]],
    Callable[[U, Q, Q, Q, V, Q, Q, X, Y, Z], UT],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        None,
        None,
        None,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, V, X, Y, PintU], Z]],
    Callable[[U, V, X, Y, Q], Z],
]:
    ...


@overload
def verify_units(
    ret: None,
    args: tuple[
        None,
        None,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
        UnitDefinition,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, V, PintU, PintV, PintX, PintZ], Z]],
    Callable[[U, V, Q, Q, Q, Q], Z],
]:
    ...


@overload
def verify_units(
    ret: ReturnUnitDefinition,
    args: tuple[
        None,
        None,
        None,
        None,
    ],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[
    [Callable[[U, V, X, Y], PintT]],
    Callable[[U, V, X, Y], Q],
]:
    ...


# The do nothing case. Here the decorator does not alter the signature of the
# input function.
@overload
def verify_units(
    ret: None,
    args: tuple[None, ...],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[FuncT], FuncT]:
    ...


def verify_units(
    ret: VerifyUnitsSupported | Iterable[VerifyUnitsSupported],
    args: Iterable[VerifyUnitsSupported],
    strict: bool = True,
    ureg: pint.registry.UnitRegistry | None = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Wrap a function to make it pint-aware

    This wraps :func:`unit_registry.wraps`, but provides better handling for types

    Parameters
    ----------
    ret
        Units of each of the return values. Use `None` to skip argument conversion.

    args
        Units of each of the arguments. Use `None` to skip argument conversion.

    strict
        Indicates that only quantities are accepted. (Default value = True)

    ureg
        Unit registry to use

        Defaults to :attr:`unit_registry` in the `fgen.units` module.

    Returns
    -------
        Decorator for wrapping callables
    """
    # Work around quirk where wraps doesn't accept dimensionless
    clean_args = [arg if arg != "dimensionless" else "" for arg in args]

    if ureg is None:
        ureg: pint.registry.UnitRegistry = pint.get_application_registry()  # type: ignore

    return ureg.wraps(ret, clean_args, strict)  # type: ignore



"""
An opinionated framework for wrapping Fortran-based code with Python.
"""
import importlib.metadata

from loguru import logger

logger.disable(__name__)

__version__ = importlib.metadata.version("fgen")


def _verify_dependency_importable(dependency: str) -> None:
    """
    Check if a package is importable

    Guards against use of :mod:`fgen`
    without first installing the template dependencies.
    This is a temporary workaround until fgen and fgen_runtime are split.
    """
    try:
        importlib.import_module(dependency)
    except ModuleNotFoundError as e:  # pragma: no cover
        raise ImportError(  # noqa: TRY003
            f"{dependency} is required. Run 'pip install fgen[templates]'"
        ) from e


_verify_dependency_importable("jinja2")
_verify_dependency_importable("black")



"""
Timeseries collection handling class
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from attrs import define

if TYPE_CHECKING:
    import pint


class ValuesBoundedLike(Protocol):
    """
    Class that behaves like a bounded values holder
    """

    @property
    def values(self) -> pint.UnitRegistry.Quantity:
        """
        Values
        """
        ...

    @property
    def value_last_bound(self) -> pint.UnitRegistry.Quantity:
        """
        Value at the end of the last bound
        """
        ...


class TimeseriesLike(Protocol):
    """
    Class that behaves like a timeseries
    """

    @property
    def name(self) -> str:
        """
        Name of the timeseries
        """
        ...

    @property
    def values(self) -> ValuesBoundedLike:
        """
        Values of the timeseries
        """
        ...

    @property
    def time(self) -> ValuesBoundedLike:
        """
        Time axis of the timeseries
        """
        ...

    @property
    def spline(self) -> int:
        """
        Spline of the timeseries
        """
        ...


@define
class TimeseriesCollection:
    """
    Container for a number of :obj:`Timeseries`
    """

    ts: list[TimeseriesLike]

    def __getitem__(self, key: int | str) -> TimeseriesLike:
        if isinstance(key, str):
            raise NotImplementedError()

        return self.ts[key]

    def __len__(self) -> int:
        return len(self.ts)



"""
Testing tools
"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Callable, Protocol

import pint.testing

from fgen_runtime.timeseries_collection import TimeseriesCollection

if TYPE_CHECKING:
    import pint


def assert_timeseries_collection_equal(
    left: TimeseriesCollection, right: TimeseriesCollection
) -> None:
    """
    Assert that two :obj:`TimeseriesCollection` are equal

    Parameters
    ----------
    left
        First object to compare

    right
        Second object to compare

    Raises
    ------
    AssertionError
        The two objects are not equal
    """

    def get_msg(i: int) -> str:
        return f"left_v.data != right_v.data for index={i}"

    assert_timeseries_collection_compare(
        left=left,
        right=right,
        compare_data=pint.testing.assert_equal,  # type: ignore # mypy and pint fighting
        get_data_error_message=get_msg,
    )


def assert_timeseries_collection_allclose(
    left: TimeseriesCollection,
    right: TimeseriesCollection,
    rtol: float = 1e-6,
    atol: float = 1e-8,
) -> None:
    """
    Assert that two :obj:`TimeseriesCollection`'s values are all close

    Parameters
    ----------
    left
        First object to compare

    right
        Second object to compare

    rtol
        Relative tolerance to pass through to :func:`pint.testing.assert_allclose`

    atol
        Absolute tolerance to pass through to :func:`pint.testing.assert_allclose`

    Raises
    ------
    AssertionError
        The two objects are not equal
    """

    def get_msg(i: int) -> str:
        return f"left_v.data not close to right_v.data for index={i}"

    assert_timeseries_collection_compare(
        left=left,
        right=right,
        compare_data=partial(pint.testing.assert_allclose, rtol=rtol, atol=atol),
        get_data_error_message=get_msg,
    )


class CompareDataCallable(Protocol):
    def __call__(
        self,
        left: pint.UnitRegistry.Quantity,
        right: pint.UnitRegistry.Quantity,
        msg: str | None = None,
    ) -> None:
        """
        Compare the data
        """
        ...


def assert_timeseries_collection_compare(
    left: TimeseriesCollection,
    right: TimeseriesCollection,
    compare_data: CompareDataCallable,
    get_data_error_message: Callable[[int], str],
) -> None:
    """
    Assert that two :obj:`TimeseriesCollection`'s values are all close

    Parameters
    ----------
    left
        First object to compare

    right
        Second object to compare

    compare_data
        Function to use to compare the data attributes of left and right

    get_data_error_message
        Function that creates the error message if the data attributes are
        not identical based on the index that is being compared.

    Raises
    ------
    AssertionError
        The two objects are not equal
    """
    # First check that there are the same number of Timeseries in each
    if len(left) != len(right):
        msg = f"{len(left)} != {len(right)}"
        raise AssertionError(msg)

    # Order matters for timeseries collections, so we can check with basic iteration
    for i in range(len(left)):  # I would want to use zip here but mypy was complaining
        left_v = left[i]
        right_v = right[i]
        if left_v.name != right_v.name:
            msg = f"{left_v.name} != {right_v.name} for index={i}"
            raise AssertionError(msg)

        msg = get_data_error_message(i)
        compare_data(
            left_v.values.values,
            right_v.values.values,
            msg=msg,
        )
        compare_data(
            left_v.values.value_last_bound,
            right_v.values.value_last_bound,
            msg=msg,
        )

        try:
            compare_data(left_v.time.values, right_v.time.values)
            compare_data(left_v.time.value_last_bound, right_v.time.value_last_bound)
        except AssertionError as exc:
            msg = f"left_v.time_axis != right_v.time_axis for index={i}"
            raise AssertionError(msg) from exc

        if left_v.spline != right_v.spline:
            msg = f"{left_v.spline} != {right_v.spline} for index={i}"
            raise AssertionError(msg)



"""
Data model of module requirements
"""

from __future__ import annotations

from attrs import define


@define
class ModuleRequirement:
    """
    Data model of a module requirement

    These specifications are used to define the requirements of a module,
    i.e. the other modules it depends on.

    We don't currently validate these at generation-time.
    If they are wrong, the module will fail to compile
    or, in the case of an incorrect :attr:`~python_module`,
    will raise an ImportError at run-time.
    """

    provides: str
    """
    Name of the derived type defined in the required module
    """

    fortran_module: str
    """
    Name of the required fortran module

    This should point to the fortran module
    that contains the definition of the required derived type.
    """

    python_module: str
    """
    Name of the python module that exposes the Python wrapper of the derived type

    This module should declare a class
    that has a name equal to the value of :attr:`~provides`.
    """



"""
Serialisation of our data models
"""
from __future__ import annotations

import copy
from pathlib import Path
from typing import Any, TypedDict, TypeVar, Union, cast

from attrs import fields
from cattrs.preconf.pyyaml import make_converter
from typing_extensions import TypeAlias

from fgen.data_models.method import Method
from fgen.data_models.module import Module
from fgen.data_models.module_enum_defining import (
    EnumDefinition,
    ModuleEnumDefining,
)
from fgen.data_models.multi_return import MultiReturn
from fgen.data_models.value import Value

T = TypeVar("T")
converter = make_converter(
    detailed_validation=False, forbid_extra_keys=True, prefer_attrib_converters=True
)
UnstructuredElement: TypeAlias = dict[
    str, Union[str, int, float, list[Any], tuple[Any, ...]]
]
UnstructuredObject: TypeAlias = Union[
    UnstructuredElement, dict[str, "UnstructuredObject"]
]


def _add_name(
    inp: dict[str, UnstructuredObject], cls: type[dict[str, T]]
) -> dict[str, T]:
    """
    Add the name to attributes and methods automatically

    This avoids having to write schemas like

    .. ::code-block

    Attributes
    ----------
          k:
            name: k
            description: something

    Methods
    -------
          calculate:
            name: calculate
            description: something else

    Instead we can just write

    .. ::code-block

    Attributes
    ----------
          k:
            description: something

    Methods
    -------
          calculate:
            description: something else

    Where the name is inferred automatically

    Parameters
    ----------
    inp
        Unstructured object

    cls
        Type to which to structure the object

    Returns
    -------
        Structured object
    """
    value_type = cls.__args__[1]  # type: ignore

    res = {}

    if inp is None:
        raise ValueError(  # noqa: TRY003  # pragma: no cover
            f"Unexpected None when structuring {cls}"
        )

    for k, v in inp.items():
        for f in fields(value_type):
            if str(f.type).startswith("dict") and v.get(f.name) is None:
                raise ValueError(  # noqa: TRY003  # pragma: no cover
                    f"{f.name!r} in {k!r} is None but a dict was expected: {v}"
                )
        if "name" in v:
            if v["name"] != k:
                raise ValueError(  # noqa: TRY003
                    f"Inconsistent name for value: {k!r} and {v['name']!r}"
                )

        if value_type == Method:
            res[k] = converter.structure({"name": k} | v, value_type)
        else:
            to_structure = cast(dict[str, dict[str, str]], copy.deepcopy(v))
            to_structure["definition"]["name"] = k
            res[k] = converter.structure(to_structure, value_type)

    return res


converter.register_structure_hook_func(
    lambda t: any(
        t == add_name_type
        for add_name_type in (
            dict[str, Value],
            dict[str, MultiReturn],
            dict[str, Method],
        )
    ),
    _add_name,
)


def _infer_return_type(inp: UnstructuredObject, _: Any) -> Union[Value, MultiReturn]:
    if "unit" not in inp:
        return converter.structure(inp, Value)

    elif isinstance(inp["unit"], str):
        return converter.structure(inp, Value)

    elif isinstance(inp["unit"], (list, tuple)):
        return converter.structure(inp, MultiReturn)

    raise NotImplementedError(inp["unit"])


converter.register_structure_hook_func(
    lambda t: any(
        t == return_requires_inference_type
        for return_requires_inference_type in [
            Union[Value, MultiReturn],
        ]
    ),
    _infer_return_type,
)


class UnstructuredEnumValue(TypedDict):
    """Unstructured enum value type hint"""

    integer_value: int
    description: str


class UnstructuredEnumDefinition(TypedDict):
    """Unstructured enum definition type hint"""

    name: str
    description: str
    values: dict[str, UnstructuredEnumValue]


class UnstructuredEnumDefiningModule(TypedDict):
    """Unstructured enum defining module"""

    name: str
    description: str
    provides: UnstructuredEnumDefinition


def _structure_enum_defining_module(
    inp: UnstructuredEnumDefiningModule,
    target_type: type[ModuleEnumDefining],
) -> ModuleEnumDefining:
    injected_values = []
    for k, v in inp["provides"]["values"].items():
        injected_values.append({"str_value": k} | v)

    injected_provides = inp["provides"] | {"values": tuple(injected_values)}
    provides = converter.structure(injected_provides, EnumDefinition)

    return ModuleEnumDefining(
        name=inp["name"],
        description=inp["description"],
        provides=provides,
    )


converter.register_structure_hook_func(
    lambda t: t == ModuleEnumDefining,
    _structure_enum_defining_module,
)


def load_module_definition(filename: str) -> Module:
    """
    Read a YAML module definition file

    This module definition contains a description of the Fortran module
    that is being wrapped.

    Parameters
    ----------
    filename
        Filename to read

    Returns
    -------
        Loaded module definition
    """
    with open(filename, encoding="utf-8") as fh:
        txt = fh.read()

    return converter.loads(txt, Module)


def load_enum_defining_module(file: Path) -> ModuleEnumDefining:
    """
    Read a YAML enum defining module definition file

    This enum defining module file contains a description of a Fortran module
    that exposes an enum.

    Parameters
    ----------
    filename
        Filename to read

    Returns
    -------
        Loaded definition of the enum defining module
    """
    with open(file, encoding="utf-8") as fh:
        txt = fh.read()

    return converter.loads(txt, ModuleEnumDefining)



"""
Data model of a Fortran derived type
"""
from __future__ import annotations

from typing import Any, Optional, Union

from attrs import define, field

from fgen.data_models.method import Method
from fgen.data_models.multi_return import MultiReturn
from fgen.data_models.value import Value


@define
class FortranDerivedType:
    """
    Data model of a Fortran derived type
    """

    name: str
    """Name of the derived type"""

    description: str
    """Description of the derived type"""

    attributes: dict[str, Value] = field()
    """The derived type's attributes"""

    methods: dict[str, Method]
    """The derived type's methods"""

    # We may need the equivalent of this for methods.
    # I haven't thought through whether it makes sense to have a method
    # that requires a Fortran unit holder without any attribute
    # that has that requirement.
    @attributes.validator
    def _check_units_handling(self, attribute: Any, value: dict[str, Value]) -> None:
        attributes_requiring_fortran_units_holder = (
            self.fortran_units_holder_reliant_attributes
        )
        if attributes_requiring_fortran_units_holder:
            if self.fortran_units_holder is None:
                msg = (
                    "There are attributes with dynamic units, "
                    "but no attribute is specified as a ``fortran_units_holder``. "
                    "Attributes with dynamic units "
                    "that require a Fortran units holder: "
                    f"{attributes_requiring_fortran_units_holder}. "
                    f"{self=} {attribute=} {value=}"
                )
                raise ValueError(msg)

    @attributes.validator
    def _check_fortran_units_holder(
        self, attribute: Any, value: dict[str, Value]
    ) -> None:
        fortran_units_holders = [
            name
            for name, att in self.attributes.items()
            if att.definition.is_fortran_units_holder
        ]
        if not fortran_units_holders:
            # No Fortran units holders, fine here
            # (checked in _check_units_handling instead).
            return

        if len(fortran_units_holders) > 1:
            # Validation should prevent getting here, just in case
            msg = (
                "Cannot have more than one Fortran units holder. "
                "These attributes have `is_fortran_units_holder=True`: "
                f"{fortran_units_holders}"
            )
            raise ValueError(msg)

        fortran_units_holder = fortran_units_holders[0]
        if not self.attributes[fortran_units_holder].definition.expose_getter_to_python:
            msg = (
                "The Fortran units holder's getter must be exposed to Python. "
                f"This is not the case for `{fortran_units_holder}`"
            )
            raise ValueError(msg)

    @property
    def fortran_units_holder_reliant_attributes(self) -> tuple[str, ...]:
        """
        The attributes whose units are reliant on ``self.fortran_units_holder``

        Fortran doesn't have an equivalent of :mod:`pint`,
        hence units are just held as plain attributes instead.
        The units of these attributes is defined by the unit holding attribute.

        Returns
        -------
            Attributes that rely on the Fortran units holder to define their units.
        """
        return tuple(
            name
            for name, att in self.attributes.items()
            if att.dynamic_unit and not isinstance(att.dynamic_unit, str)
        )

    @property
    def fortran_units_holder(self) -> Union[str, None]:
        """
        The attribute on the Fortran derived type which holds the units

        This is often, but does not necessarily need to be, an attribute named `units`.

        Fortran doesn't have an equivalent of :mod:`pint`,
        hence units are just held as plain attributes instead.
        If a provided derived type has such an attribute,
        it can be accessed via this property

        Returns
        -------
            Attribute on the Fortran derived type which holds the units.
            If no attribute on the Fortran derived type holds units,
            ``None`` is returned.
        """
        fortran_units_holders = [
            name
            for name, att in self.attributes.items()
            if att.definition.is_fortran_units_holder
        ]
        if not fortran_units_holders:
            return None

        if len(fortran_units_holders) > 1:  # pragma: no cover
            # Validation should prevent getting here, just in case
            msg = (
                "Validation should have prevented this. "
                "Cannot have more than one Fortran units holder. "
                "These attributes have `is_fortran_units_holder=True`: "
                f"{fortran_units_holders}"
            )
            raise AssertionError(msg)

        return fortran_units_holders[0]

    @property
    def exposed_attributes(self) -> dict[str, Value]:
        """
        Get the attributes that are marked to be exposed to python

        Returns
        -------
            Collection of exposed attributes
        """
        return dict(
            filter(
                lambda item: item[1].definition.expose_getter_to_python,
                self.attributes.items(),
            )
        )

    @property
    def units(self) -> dict[str, Union[str, tuple[str, ...]]]:
        """
        Get the unit for each declared value in the derived type

        The unit for a given named value must be consistent across the derived type
        and the module which defines it.
        This includes the unit for ``attributes``, and in the ``parameter`` and
        ``return`` values for each ``method``.

        Raises
        ------
        ValueError
            Inconsistent units were found

        Returns
        -------
            Dictionary containing value names' as keys and the associated units
            as values.
        """
        # We are tracking the units for each possible source of units.
        # For each name we keep track of a tuple containing the method name
        # where the unit was declared and the unit.
        # For derived type attributes the "method name" is None.

        # Collect units from the attributes.
        unit_sources: dict[str, tuple[Optional[str], Union[str, tuple[str, ...]]]] = {
            key: (None, attr.unit)
            for key, attr in self.attributes.items()
            # Drop any attributes that don't have a unit
            if attr.unit is not None
        }

        # Collect units from the methods
        for method_name, method in self.methods.items():
            method_units = method.units

            # Check that the method units are consistent
            # with any previously declared units
            for name, method_unit in method_units.items():
                if name in unit_sources and unit_sources[name][1] != method_unit:
                    # Inconsistent unit declarations found

                    previous_source, previous_value = unit_sources[name]
                    if previous_source is not None:
                        previous_source_type = "method"
                    else:
                        previous_source_type = "derived type"
                        previous_source = self.name

                    raise ValueError(  # noqa: TRY003
                        f"Inconsistent units for attribute '{name}'. "
                        f"In the method '{method_name}' it has units '{method_unit}' "
                        f"whereas in the {previous_source_type} '{previous_source}' "
                        f"it has units '{previous_value}'"
                    )
                unit_sources[name] = (method_name, method_unit)

        return {key: value[1] for key, value in unit_sources.items()}

    @property
    def units_multi_return(self) -> dict[str, tuple[str, ...]]:
        """
        Get units declared in this object, if the units are tuples

        This occurs if the units apply to a return statement that returns
        multiple values

        Returns
        -------
            Dictionary containing value names' as keys and the associated units
            as values (all these values are plain strings).
        """
        return {k: v for k, v in self.units.items() if not isinstance(v, str)}

    def get_dynamic_unit_source(
        self, value: Union[MultiReturn, Value]
    ) -> Optional[str]:
        """
        Get the source of the dynamic unit for a given value

        Parameters
        ----------
        value
            Value for which to get the source of the dynamic unit

        Returns
        -------
            Source of the dynamic unit for ``value``
        """
        if not value.dynamic_unit:
            return None

        if isinstance(value.dynamic_unit, str):
            return value.dynamic_unit

        # If we have a dynamic unit, but it isn't defined as a string,
        # then it comes from the Fortran units holder attribute of self.
        return f"self.{self.fortran_units_holder}"



"""
Helpers for type annotations
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from fgen.data_models import MultiReturn, Value


PINT_QUANTITY_TYPE_ANNOTATION: str = "pint.registry.UnitRegistry.Quantity"
"""
Type annotation for pint quantities.

Currently these don't support information about the type held by the quantity.
This can be updated in future.
"""


NUMPY_ARRAY_TYPE_ANNOTATION_BASE: str = "npt.NDArray[{kind}]"
"""
Base type annotation for numpy arrays
"""


def get_numpy_array_type_annotation(value: Union[MultiReturn, Value]) -> str:
    """
    Get type annotation for a numpy array
    """
    fdt = value.definition.as_fortran_data_type()
    if fdt.is_float_double:
        return NUMPY_ARRAY_TYPE_ANNOTATION_BASE.format(kind="np.float64")

    raise NotImplementedError(value)
